import os
from yggdrasil import tools
from yggdrasil.runner import run
tyk_dir = 'tyk2'

try:
    # Change to the solution directory
    old_dir = os.getcwd()
    if not old_dir.endswith(tyk_dir):
        os.chdir(os.path.join('solutions', tyk_dir))
    if not os.path.isdir('output'):
        os.mkdir('output')

    # Part 1: Add interface calls
    tools.display_source_diff('../../models/co2.py', 'models/co2.py', number_lines=True)
    run(['yamls/co2.yml'], production_run=True)

    # Part 2: Write a YAML
    tools.display_source('yamls/humidity.yml', number_lines=True)
    run(['yamls/humidity.yml'], production_run=True)

    # Part 3: Add interface calls and write a YAML
    tools.display_source_diff('../../models/reflectance.py', 'models/reflectance.py', number_lines=True)
    tools.display_source('yamls/reflectance.yml', number_lines=True)
    run(['yamls/reflectance.yml'], production_run=True)

finally:
    os.chdir(old_dir)



import pytest
from tests.metaschema.datatypes.test_MetaschemaType import (
    TestMetaschemaType as base_class)
import copy
import numpy as np


def test_coerce(nested_approx):
    r"""Test serialization of coerced types."""
    from yggdrasil import serialize
    from yggdrasil.metaschema.datatypes.JSONArrayMetaschemaType import (
        JSONArrayMetaschemaType)
    typedef = {'type': 'array', 'items': [{'type': '1darray',
                                           'subtype': 'float',
                                           'title': 'a',
                                           'precision': 64}]}
    x = JSONArrayMetaschemaType(**typedef)
    key_order = ['a']
    msg_recv = [np.zeros(3, 'float64')]
    msg_send_list = [msg_recv[0],
                     np.zeros(3, 'float32'),
                     serialize.list2numpy(msg_recv, names=key_order),
                     serialize.list2pandas(msg_recv, names=key_order),
                     serialize.list2dict(msg_recv, names=key_order)]

    def do_send_recv(msg_send):
        msg_seri = x.serialize(msg_send, tyepdef=typedef, key_order=key_order)
        assert(x.deserialize(msg_seri)[0] == nested_approx(msg_recv))

    for y in msg_send_list:
        do_send_recv(y)
    assert(JSONArrayMetaschemaType.coerce_type({'a': 'hello', 'b': 'world'})
           == [{'a': 'hello', 'b': 'world'}])
    

class TestJSONArrayMetaschemaType(base_class):
    r"""Test class for JSONArrayMetaschemaType class with float."""

    _mod = 'yggdrasil.metaschema.datatypes.JSONArrayMetaschemaType'
    _cls = 'JSONArrayMetaschemaType'

    @pytest.fixture(scope="class")
    def value(self, container_values):
        r"""list: Test value."""
        return [v for v in container_values]

    @pytest.fixture(scope="class")
    def fulldef(self, python_class, container_definitions):
        r"""dict: Full type definitions."""
        return {'type': python_class.name,
                'items': [v for v in container_definitions]}

    @pytest.fixture(scope="class")
    def typedef_base(self, container_typedefs):
        r"""dict: Base type definition."""
        return {'items': [v for v in container_typedefs]}
        
    @pytest.fixture(scope="class")
    def valid_encoded(self, fulldef):
        r"""list: Encoded objects that are valid under this type."""
        return [fulldef]

    @pytest.fixture(scope="class")
    def valid_decoded(self, value):
        r"""list: Objects that are valid under this type."""
        return [value]

    @pytest.fixture(scope="class")
    def invalid_encoded(self, fulldef):
        r"""list: Encoded objects that are invalid under this type."""
        out = [{},
               {'type': fulldef['type'],
                'items': [fulldef['items'][0]]},
               copy.deepcopy(fulldef)]
        del out[-1]['items'][0]['type']
        out.append(copy.deepcopy(fulldef))
        out[-1]['items'][0]['type'] = 'invalid'
        return out
    
    @pytest.fixture(scope="class")
    def compatible_objects(self, value):
        r"""list: Objects that are compatible with this type."""
        return [(value, value, None)]
        
    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return [(None, None), ('1, 1 ', ['1', '1'])]

    def test_encode_data_readable(self, python_class):
        r"""Test corner case of encode_data_readable."""
        python_class.encode_data_readable(['1', '1'], {})

    def test_container_errors(self, python_class):
        r"""Test errors on container operations."""
        with pytest.raises(RuntimeError):
            python_class._assign([], 10, None)

    def test_item_dictionary(self, python_class):
        r"""Test dictionary as items value."""
        x = [1, 2, 3]
        typedef = {'type': 'array', 'items': {'type': 'int'}}
        python_class.validate_instance(x, typedef)
        python_class.encode_data(x, typedef)

    def test_validate_errors(self, python_class):
        r"""Test error on validation of non-structured array."""
        with pytest.raises(ValueError):
            python_class.validate(np.zeros(5), raise_errors=True)



import os
from yggdrasil.runner import run
from yggdrasil import tools
tyk_dir = 'tyk8'

try:
    # Change to the solution directory
    old_dir = os.getcwd()
    if not old_dir.endswith(tyk_dir):
        os.chdir(os.path.join('solutions', tyk_dir))
    if not os.path.isdir('output'):
        os.mkdir('output')

    # Part 1: Sort intensities
    tools.display_source_diff('../../yamls/connections_v0.yml',
                              'yamls/connections_sort_intensity.yml',
                              number_lines=True)
    run(['yamls/light_v0_python.yml', 'yamls/connections_sort_intensity.yml'],
        production_run=True)
    tools.display_source('output/light_filter_lt400.txt', number_lines=True)
    tools.display_source('output/light_filter_ge400.txt', number_lines=True)

    # Part 2: New light model
    tools.display_source_diff('models/light_v0.py', 'models/light_new.py',
                              number_lines=True)
    tools.display_source_diff('yamls/light_v0_python.yml',
                              'yamls/light_new_python.yml', number_lines=True)
    tools.display_source_diff('../../yamls/connections_v0.yml',
                              'yamls/connections_sort_height.yml',
                              number_lines=True)
    run(['yamls/light_v0_python.yml', 'yamls/light_new_python.yml',
         'yamls/connections_sort_height.yml'],
        production_run=True)
    tools.display_source('output/light_multi_model.txt', number_lines=True)

    # Part 3: Transformation
    tools.display_source_diff('yamls/connections_sort_height.yml',
                              'yamls/connections_sort_height_transform.yml',
                              number_lines=True)
    run(['yamls/light_v0_python.yml', 'yamls/light_new_python.yml',
         'yamls/connections_sort_height_transform.yml'],
        production_run=True)
    tools.display_source('output/light_multi_model_transform.txt',
                         number_lines=True)

finally:
    os.chdir(old_dir)



import pytest
import copy
import jsonschema
from yggdrasil import constants
from yggdrasil.metaschema import MetaschemaTypeError
from tests import TestClassBase as base_class


class TestMetaschemaType(base_class):
    r"""Test class for MetaschemaType class."""

    _mod = 'yggdrasil.metaschema.datatypes.MetaschemaType'
    _cls = 'MetaschemaType'

    @pytest.fixture(scope="class")
    def explicit(self):
        r"""bool: If True the type is explicit."""
        return False

    @pytest.fixture(scope="class")
    def empty_msg(self):
        r"""object: Empty object for testing."""
        return b''

    @pytest.fixture(scope="class")
    def typedef_base(self):
        r"""dict: Base type definition."""
        return {}

    @pytest.fixture(scope="class")
    def valid_encoded(self):
        r"""list: Encoded objects that are valid under this type."""
        return []

    @pytest.fixture(scope="class")
    def valid_decoded(self):
        r"""list: Objects that are valid under this type."""
        return ['nothing']

    @pytest.fixture(scope="class")
    def invalid_encoded(self):
        r"""list: Encoded objects that are invalid under this type."""
        return [{}]

    @pytest.fixture(scope="class")
    def invalid_decoded(self):
        r"""list: Objects that are invalid under this type."""
        return []

    @pytest.fixture(scope="class")
    def invalid_validate(self):
        r"""list: Objects that are invalid under this type."""
        return [None]

    @pytest.fixture(scope="class")
    def compatible_objects(self):
        r"""list: Objects that are compatible with this type."""
        return []

    @pytest.fixture(scope="class")
    def encode_type_kwargs(self):
        r"""dict: Keyword arguments for encoding this type."""
        return {}

    @pytest.fixture(scope="class")
    def encode_data_kwargs(self):
        r"""dict: Keyword arguments for encoding data of this type."""
        return {}

    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return [(None, None)]

    @pytest.fixture(scope="class")
    def typedef(self, python_class, typedef_base):
        r"""dict: Type definition"""
        return dict(typedef_base, type=python_class.name)

    @pytest.fixture
    def instance_kwargs(self, typedef_base):
        r"""Keyword arguments for a new instance of the tested class."""
        return typedef_base

    @pytest.fixture
    def nested_result(self, nested_approx):
        r"""Prepare value for comparison."""
        def nested_result_w(x):
            return nested_approx(x)
        return nested_result_w

    def test_generate_data(self, class_name, python_class, valid_encoded):
        r"""Test generation of data."""
        if class_name == 'MetaschemaType':
            return
        if len(valid_encoded) > 0:
            typedef = valid_encoded[0]
            data = python_class.generate_data(typedef)
            python_class.validate(data, raise_errors=True)

    def test_issubtype(self, class_name, python_class, typedef):
        r"""Test issubtype."""
        if class_name == 'MetaschemaType':
            return
        assert(python_class.issubtype(typedef['type']))
        if python_class.name == 'any':
            assert(python_class.issubtype('invalid'))
        else:
            assert(not python_class.issubtype('invalid'))
        if not isinstance(typedef['type'], (list, tuple)):
            assert(python_class.issubtype([typedef['type']]))

    def test_validate(self, class_name, python_class, valid_decoded,
                      invalid_validate):
        r"""Test validation."""
        if class_name == 'MetaschemaType':
            for x in valid_decoded:
                with pytest.raises(NotImplementedError):
                    python_class.validate(x)
        else:
            for x in valid_decoded:
                assert(python_class.validate(x) is True)
            for x in invalid_validate:
                assert(python_class.validate(x) is False)
                with pytest.raises(BaseException):
                    python_class.validate(x, raise_errors=True)

    def test_normalize(self, python_class, nested_result, valid_normalize):
        r"""Test normalization."""
        for (x, y) in valid_normalize:
            z = python_class.normalize(x)
            assert(z == nested_result(y))

    def test_fixed2base(self, python_class, explicit, typedef):
        r"""Test conversion of type definition from fixed type to the base."""
        if explicit:
            t1 = copy.deepcopy(typedef)
            x1 = python_class.typedef_fixed2base(t1)
            t2 = copy.deepcopy(x1)
            t2['type'] = t1['type']
            x2 = python_class.typedef_fixed2base(t2)
            assert(x1 == x2)
            y = python_class.typedef_base2fixed(x1)
            assert(y == typedef)

    def test_extract_typedef(self, python_class, valid_encoded):
        r"""Test extract_typedef."""
        if len(valid_encoded) > 0:
            python_class.extract_typedef(valid_encoded[0])

    def test_update_typedef(self, python_class, instance, typedef,
                            explicit):
        r"""Test update_typedef raises error on non-matching typename."""
        instance.update_typedef(**typedef)
        with pytest.raises(MetaschemaTypeError):
            instance.update_typedef(type='invalid')
        if explicit:
            typedef_base = python_class.typedef_fixed2base(typedef)
            instance.update_typedef(**typedef_base)

    def test_definition_schema(self, python_class):
        r"""Test definition schema."""
        s = python_class.definition_schema()
        # jsonschema.Draft3Validator.check_schema(s)
        jsonschema.Draft4Validator.check_schema(s)

    def test_metadata_schema(self, python_class):
        r"""Test metadata schema."""
        s = python_class.metadata_schema()
        # jsonschema.Draft3Validator.check_schema(s)
        jsonschema.Draft4Validator.check_schema(s)

    def test_encode_data(self, python_class, valid_decoded, typedef,
                         encode_type_kwargs, encode_data_kwargs,
                         nested_result):
        r"""Test encode/decode data & type."""
        if python_class.__name__ == 'MetaschemaType':
            for x in valid_decoded:
                with pytest.raises(NotImplementedError):
                    python_class.encode_type(x)
                with pytest.raises(NotImplementedError):
                    python_class.encode_data(x, typedef)
            with pytest.raises(NotImplementedError):
                python_class.decode_data(None, typedef)
        else:
            for x in valid_decoded:
                y = python_class.encode_type(x, **encode_type_kwargs)
                z = python_class.encode_data(x, y, **encode_data_kwargs)
                python_class.encode_data_readable(x, None)
                python_class.encode_data_readable(x, y)
                x2 = python_class.decode_data(z, y)
                assert(x2 == nested_result(x))
            if python_class.__name__ not in ['JSONNullMetaschemaType',
                                             'AnyMetaschemaType']:
                with pytest.raises(MetaschemaTypeError):
                    python_class.encode_type(None)

    def test_check_encoded(self, python_class, valid_encoded, invalid_encoded,
                           typedef):
        r"""Test check_encoded."""
        # Test invalid for incorrect typedef
        if len(valid_encoded) > 0:
            assert(python_class.check_encoded(valid_encoded[0], {}) is False)
            with pytest.raises(BaseException):
                python_class.check_encoded(valid_encoded[0], {}, raise_errors=True)
        # Test valid
        for x in valid_encoded:
            assert(python_class.check_encoded(x, typedef, raise_errors=True))
        # Test invalid
        for x in invalid_encoded:
            assert(python_class.check_encoded(x, typedef) is False)
            with pytest.raises(BaseException):
                python_class.check_encoded(x, typedef, raise_errors=True)

    def test_check_decoded(self, python_class, valid_decoded, typedef,
                           invalid_validate, invalid_decoded):
        r"""Test check_decoded."""
        # Not implemented for base class
        if python_class.__name__ == 'MetaschemaType':
            for x in valid_decoded:
                with pytest.raises(NotImplementedError):
                    python_class.check_decoded(x, typedef)
        else:
            # Test object alone
            if len(valid_decoded) > 0:
                x = valid_decoded[0]
                assert(python_class.check_decoded(x, None) is True)
            # Test valid
            for x in valid_decoded:
                assert(python_class.check_decoded(x, typedef) is True)
            # Test invalid with incorrect typedef
            for x in valid_decoded:
                assert(python_class.check_decoded(x, {}) is False)
                with pytest.raises(BaseException):
                    python_class.check_decoded(x, {}, raise_errors=True)
            # Test invalid
            for x in (invalid_validate + invalid_decoded):
                assert(python_class.check_decoded(x, typedef) is False)
                with pytest.raises(BaseException):
                    python_class.check_decoded(x, typedef, raise_errors=True)

    def test_encode_errors(self, python_class, invalid_validate, typedef):
        r"""Test error on encode."""
        if python_class.__name__ == 'MetaschemaType':
            if invalid_validate:
                with pytest.raises(NotImplementedError):
                    python_class.encode(invalid_validate[0], typedef)
        else:
            if invalid_validate:
                with pytest.raises((ValueError,
                                    jsonschema.exceptions.ValidationError)):
                    python_class.encode(invalid_validate[0], typedef)

    def test_decode_errors(self, python_class, invalid_encoded, typedef):
        r"""Test error on decode."""
        if invalid_encoded:
            with pytest.raises((ValueError,
                                jsonschema.exceptions.ValidationError)):
                python_class.decode(invalid_encoded[0], typedef)

    def test_transform_type(self, python_class, compatible_objects,
                            nested_result):
        r"""Test transform_type."""
        for x, y, typedef in compatible_objects:
            z = python_class.transform_type(x, typedef)
            assert(z == nested_result(y))

    def test_serialize(self, python_class, instance, valid_decoded,
                       nested_result):
        r"""Test serialize/deserialize."""
        if python_class.__name__ == 'MetaschemaType':
            for x in valid_decoded:
                with pytest.raises(NotImplementedError):
                    instance.serialize(x)
        else:
            for x in valid_decoded:
                msg = instance.serialize(x)
                y = instance.deserialize(msg)
                assert(y[0] == nested_result(x))

    def test_serialize_error(self, python_class, valid_decoded, instance):
        r"""Test serialization errors."""
        if (python_class.__name__ != 'MetaschemaType') and (len(valid_decoded) > 0):
            with pytest.raises(RuntimeError):
                instance.serialize(valid_decoded[0], data='something')

    def test_deserialize_error(self, python_class, instance, valid_decoded):
        r"""Test deserialization errors."""
        with pytest.raises(TypeError):
            instance.deserialize(self)
        with pytest.raises(ValueError):
            instance.deserialize(b'invalid')
        if (python_class.__name__ != 'MetaschemaType') and (len(valid_decoded) > 0):
            with pytest.raises(ValueError):
                instance.deserialize(instance.serialize(valid_decoded[0]),
                                     metadata={'size': 0})
        
    def test_deserialize_empty(self, instance, empty_msg, nested_result):
        r"""Test call for empty string."""
        # Completely empty
        out = instance.deserialize(empty_msg)
        assert(out[0] == nested_result(instance._empty_msg))
        assert(out[1] == dict(size=0, incomplete=False))
        # Empty metadata and message
        out = instance.deserialize((2 * constants.YGG_MSG_HEAD) + empty_msg)
        assert(out[0] == nested_result(instance._empty_msg))
        assert(out[1] == dict(size=0, incomplete=False))

    def test_deserialize_incomplete(self, python_class, valid_decoded,
                                    instance):
        r"""Test call for incomplete message."""
        if (python_class.__name__ != 'MetaschemaType') and (len(valid_decoded) > 0):
            out = instance.serialize(valid_decoded[0])
            obj, metadata = instance.deserialize(out[:-1])
            assert(metadata['incomplete'] is True)



from yggdrasil.tools import pprint_encoded
from yggdrasil.examples.transforms import get_test_data


def src():
    test_data = get_test_data()
    print('SENDING')
    pprint_encoded(test_data)
    return test_data



from yggdrasil import multitasking
from yggdrasil.communication import CommBase, NoMessages
from yggdrasil.config import ygg_cfg
import logging
logger = logging.getLogger(__name__)
try:
    import pika
    _rmq_installed = True
    _pika_version_maj = int(float(pika.__version__.split('.')[0]))
    if _pika_version_maj < 1:  # pragma: debug
        raise ImportError("pika version <1.0 no longer supported.")
except ImportError:
    logger.debug("Could not import pika. "
                 + "RabbitMQ support will be disabled.")
    pika = None
    _rmq_installed = False
    _pika_version_maj = 0


_rmq_param_sep = '_RMQPARAM_'
_localhost = '127.0.0.1'


def get_rmq_parameters(url=None, user=None, username=None,
                       password=None, host=None, virtual_host=None,
                       vhost=None, port=None, exchange=None, queue=''):
    r"""Get RabbitMQ connection parameters.

    Args:
        url (str, optional): Address of RMQ server that includes all the
            necessary information. If this is provided, the remaining arguments
            are ignored. Defaults to None and the connection parameters are
            taken from the other arguments.
        user (str, optional): RabbitMQ server username. Defaults to config
            option 'user' in section 'rmq' if it exists and 'guest' if it
            does not.
        username (str, optional): Alias for user.
        password (str, optional): RabbitMQ server password. Defaults to
            config option 'password' in section 'rmq' if it exists and
            'guest' if it does not.
        host (str, optional): RabbitMQ server host. Defaults to config option
            'host' in section 'rmq' if it exists and _localhost if it
            does not. If _localhost, the output of socket.gethostname()
            is used.
        virtual_host (str, optional): RabbitMQ server virtual host. Defaults
            to config option 'vhost' in section 'rmq' if it exists and '/'
            if it does not.
        vhost (str, optional): Alias for virtual_host.
        port (str, optional): Port on host to use. Defaults to config option
            'port' in section 'rmq' if it exists and '5672' if it does not.
        exchange (str, optional): RabbitMQ exchange. Defaults to config
            option 'namespace' in section 'rmq' if it exits and '' if it does
            not.
        queue (str, optional): Name of the queue that messages will be
            send to or received from. If an empty string, the queue will
            be a random string and exclusive to a receiving comm. Defaults
            to ''.

    Returns:
        tuple: The connection url, exchange, & queue.

    """
    if url is None:
        if user is None:
            user = username or ygg_cfg.get('rmq', 'user', 'guest')
        if password is None:
            password = ygg_cfg.get('rmq', 'password', 'guest')
        if host is None:
            host = ygg_cfg.get('rmq', 'host', _localhost)
        if virtual_host is None:
            virtual_host = vhost or ygg_cfg.get('rmq', 'vhost', '/')
        if virtual_host == '/':
            virtual_host = '%2f'
        if port is None:
            port = ygg_cfg.get('rmq', 'port', '5672')
        url = 'amqp://%s:%s@%s:%s/%s' % (
            user, password, host, port, virtual_host)
    if exchange is None:
        exchange = ygg_cfg.get('rmq', 'namespace', '')
    return url, exchange, queue


def check_rmq_server(url=None, **kwargs):
    r"""Check that connection to a RabbitMQ server is possible.

    Args:
        url (str, optional): Address of RMQ server that includes all the
            necessary information. If this is provided, the remaining arguments
            are ignored. Defaults to None and the connection parameters are
            taken from the other arguments.
        username (str, optional): RMQ server username. Defaults to config value.
        password (str, optional): RMQ server password. Defaults to config value.
        host (str, optional): RMQ hostname or IP Address to connect to. Defaults
            to config value.
        port (str, optional): RMQ server TCP port to connect to. Defaults to
            config value.
        vhost (str, optional): RMQ virtual host to use. Defaults to config value.

    Returns:
        bool: True if connection to RabbitMQ server is possible, False
            otherwise.

    """
    out = True
    if not _rmq_installed:
        return False
    if url is not None:
        parameters = pika.URLParameters(url)
    else:
        username = kwargs.get('username', ygg_cfg.get('rmq', 'user', 'guest'))
        password = kwargs.get('password', ygg_cfg.get('rmq', 'password', 'guest'))
        host = kwargs.get('host', ygg_cfg.get('rmq', 'host', _localhost))
        port = kwargs.get('port', ygg_cfg.get('rmq', 'port', '5672'))
        vhost = kwargs.get('vhost', ygg_cfg.get('rmq', 'vhost', '/'))
        credentials = pika.PlainCredentials(username, password)
        parameters = pika.ConnectionParameters(host=host, port=int(port),
                                               virtual_host=vhost,
                                               credentials=credentials)
    # Try to establish connection
    logging.getLogger("pika").propagate = False
    try:
        from yggdrasil import tools
        with tools.track_fds("pika test connection: "):
            connection = pika.BlockingConnection(parameters)
        if not connection.is_open:  # pragma: debug
            del connection
            raise BaseException("Connection was not opened.")
        connection.close()
        del connection
    except BaseException as e:  # pragma: debug
        print("Error when attempting to connect to the RabbitMQ server: "
              + str(e))
        out = False
    logging.getLogger("pika").propagate = True
    return out


class RMQServer(CommBase.CommServer):
    r"""RMQ server object for cleaning up server connections."""

    def __init__(self, *args, **kwargs):
        self.comm_cls = kwargs.get('comm_cls', RMQComm)
        super(RMQServer, self).__init__(*args, **kwargs)

    def terminate(self, *args, **kwargs):
        self.comm_cls.unregister_comm(self.srv_address)
        super(RMQServer, self).terminate(*args, **kwargs)


class RMQComm(CommBase.CommBase):
    r"""Class for handling basic RabbitMQ communications.

    Attributes:
        connection (:class:`pika.Connection`): RabbitMQ connection.
        channel (:class:`pika.Channel`): RabbitMQ channel.

    Raises:
        RuntimeError: If a connection cannot be established.

    Developer Notes:
        It is not advised that new language implement a RabbitMQ communication
        interface. Rather RMQ communication is included explicitly for
        connections between models that are not co-located on the same machine
        and are used by the |yggdrasil| framework connections on the Python side.

    """

    _commtype = 'rmq'
    _schema_subtype_description = ('RabbitMQ connection.')
    # Based on limit of 32bit int, this could be 2**30, but this is
    # too large for stack allocation in C so 2**20 will be used.
    _maxMsgSize = 2**20
    address_description = ("AMPQ queue address of the form "
                           "``<url>_RMQPARAM_<exchange>_RMQPARAM_<queue>`` "
                           "where ``url`` is the broker address (see explanation "
                           "`here <https://pika.readthedocs.io/en/stable/"
                           "examples/using_urlparameters.html>`_), "
                           "``exchange`` is the name of the exchange on the queue "
                           "that should be used, and ``queue`` is the name of "
                           "the queue.")
    _disconnect_attr = (CommBase.CommBase._disconnect_attr
                        + ['_opening', '_closing'])
    
    def _init_before_open(self, **kwargs):
        r"""Set null connection and channel."""
        if not hasattr(self, 'rmq_lock'):
            self.rmq_lock = multitasking.RLock()
        self.connection = None
        self.channel = None
        self._opening = multitasking.ProcessEvent()
        self._closing = multitasking.ProcessEvent()
        self._server_class = RMQServer
        self._server_kwargs = {'comm_cls': self.__class__}
        super(RMQComm, self)._init_before_open(**kwargs)

    @property
    def url(self):
        r"""str: AMQP server address."""
        return self.address.split(_rmq_param_sep)[0]

    @property
    def exchange(self):
        r"""str: AMQP exchange."""
        return self.address.split(_rmq_param_sep)[1]

    @property
    def queue(self):
        r"""str: AMQP queue."""
        return self.address.split(_rmq_param_sep)[2]

    @classmethod
    def new_comm_kwargs(cls, name, user=None, password=None, host=None,
                        virtual_host=None, port=None, exchange=None, queue='',
                        **kwargs):
        r"""Initialize communication with new connection.

        Args:
            name (str): Name of new connection.
            user (str, optional): RabbitMQ server username. Defaults to config
                option 'user' in section 'rmq' if it exists and 'guest' if it
                does not.
            password (str, optional): RabbitMQ server password. Defaults to
                config option 'password' in section 'rmq' if it exists and
                'guest' if it does not.
            host (str, optional): RabbitMQ server host. Defaults to config option
                'host' in section 'rmq' if it exists and _localhost if it
                does not. If _localhost, the output of socket.gethostname()
                is used.
            virtual_host (str, optional): RabbitMQ server virtual host. Defaults
                to config option 'vhost' in section 'rmq' if it exists and '/'
                if it does not.
            port (str, optional): Port on host to use. Defaults to config option
                'port' in section 'rmq' if it exists and '5672' if it does not.
            exchange (str, optional): RabbitMQ exchange. Defaults to config
                option 'namespace' in section 'rmq' if it exits and '' if it does
                not.
            queue (str, optional): Name of the queue that messages will be
                send to or received from. If an empty string, the queue will
                be a random string and exclusive to a receiving comm. Defaults
                to ''.
            **kwargs: Additional keywords arguments are returned as keyword
                arguments for the new comm.

        Returns:
            tuple(tuple, dict): Arguments and keyword arguments for new comm.
        
        """
        args = [name]
        if 'address' not in kwargs:
            (url, exchange, queue) = get_rmq_parameters(
                user=user, password=password, host=host,
                virtual_host=virtual_host, port=port, exchange=exchange,
                queue=queue)
            kwargs['address'] = _rmq_param_sep.join([url, exchange, queue])
        return args, kwargs

    def opp_comm_kwargs(self, for_yaml=False):
        r"""Get keyword arguments to initialize communication with opposite
        comm object.

        Args:
            for_yaml (bool, optional): If True, the returned dict will only
                contain values that can be specified in a YAML file. Defaults
                to False.

        Returns:
            dict: Keyword arguments for opposite comm object.

        """
        kwargs = super(RMQComm, self).opp_comm_kwargs(for_yaml=for_yaml)
        return kwargs

    def bind(self):
        r"""Declare queue to get random new queue."""
        if self._opening.has_started() or self._closing.has_started():
            return
        self._opening.start()
        with self.rmq_lock:
            if self.connection is None:
                parameters = pika.URLParameters(self.url)
                self.connection = pika.BlockingConnection(parameters)
                self.original_queue = self.queue
            self.channel = self.connection.channel()
            self.channel.exchange_declare(exchange=self.exchange,
                                          auto_delete=True)
            if self.direction == 'send':
                self.channel.confirm_delivery()
            res = self.channel.queue_declare(
                queue=self.queue, exclusive=False, auto_delete=True,
                passive=self.original_queue.startswith('amq.'))
            if not self.queue:
                self.address += res.method.queue
            self.channel.queue_bind(exchange=self.exchange,
                                    queue=self.queue)
            self.register_comm(self.address, (self.connection, self.channel))
            super(RMQComm, self).bind()
        self._opening.stop()

    def _close(self, linger=False):
        r"""Close the connection.

        Args:
            linger (bool, optional): If True, drain messages before closing the
                comm. Defaults to False.

        """
        if self._closing.has_started():  # pragma: debug
            return
        self._closing.start()
        if self.direction == 'recv':
            self.close_queue()
        self.close_channel()
        self.close_connection()
        if not self.is_client:
            self.unregister_comm(self.address)
        with self.rmq_lock:
            self.channel = None
            self.connection = None
        super(RMQComm, self)._close(linger=linger)
        self._closing.stop()

    def close_queue(self, skip_unbind=False):
        r"""Close the queue if the channel exists."""
        if self.direction != 'recv':
            return
        with self.rmq_lock:
            if self.channel and (not self.is_client):
                try:
                    if not skip_unbind:
                        self.channel.queue_unbind(queue=self.queue,
                                                  exchange=self.exchange)
                    self.channel.queue_delete(queue=self.queue)
                except (pika.exceptions.ChannelClosed,
                        pika.exceptions.ConnectionClosed,
                        pika.exceptions.ChannelWrongStateError,
                        pika.exceptions.ConnectionWrongStateError):  # pragma: debug
                    pass
                except AttributeError:  # pragma: debug
                    if self.channel is not None:
                        raise

    def close_channel(self):
        r"""Close the channel if it exists."""
        with self.rmq_lock:
            if self.channel is not None:
                self.debug('Closing the channel')
                try:
                    self.channel.close()
                except (pika.exceptions.ChannelWrongStateError,
                        pika.exceptions.StreamLostError):  # pragma: debug
                    pass

    def close_connection(self, *args, **kwargs):
        r"""Close the connection."""
        with self.rmq_lock:
            if self.connection is not None:
                self.debug('Closing connection')
                self.connection.close(*args, **kwargs)

    def atexit(self):  # pragma: debug
        r"""Close operations."""
        if self.direction == 'send':
            self.linger()
        super(RMQComm, self).atexit()
        
    @property
    def is_open(self):
        r"""bool: True if the connection and channel are open."""
        with self.rmq_lock:
            if self.channel is None or self.connection is None:
                return False
            if not self.connection.is_open:
                return False
            if not self.channel.is_open:
                return False
            return (self._opening.has_stopped()
                    and (not self._closing.has_started()))

    def get_queue_result(self):
        r"""Get the fram from passive queue declare."""
        with self.rmq_lock:
            res = None
            if self.is_open:
                try:
                    res = self.channel.queue_declare(queue=self.queue,
                                                     auto_delete=True,
                                                     passive=True)
                except pika.exceptions.ChannelClosedByBroker:  # pragma: debug
                    self._close()
                # except BlockingIOError:  # pragma: debug
                #     self.sleep()
                #     res = self.get_queue_result()
                # except (pika.exceptions.ChannelClosed,
                #         pika.exceptions.ConnectionClosed,
                #         pika.exceptions.ChannelWrongStateError,
                #         pika.exceptions.ConnectionWrongStateError,
                #         pika.exceptions.StreamLostError,
                #         AttributeError):  # pragma: debug
                #     self._close()
            return res
        
    @property
    def n_msg_recv(self):
        r"""int: Number of messages in the queue."""
        out = 0
        res = self.get_queue_result()
        if res is not None:
            out = res.method.message_count
        return out

    @property
    def n_msg_send(self):
        r"""int: Number of messages in the queue."""
        return 0

    @property
    def is_confirmed_send(self):
        r"""bool: True if all sent messages have been confirmed."""
        return True

    @property
    def is_confirmed_recv(self):
        r"""bool: True if all received messages have been confirmed."""
        return True

    @property
    def get_work_comm_kwargs(self):
        r"""dict: Keyword arguments for an existing work comm."""
        out = super(RMQComm, self).get_work_comm_kwargs
        out['exchange'] = self.exchange
        return out

    @property
    def create_work_comm_kwargs(self):
        r"""dict: Keyword arguments for a new work comm."""
        out = super(RMQComm, self).create_work_comm_kwargs
        out['exchange'] = self.exchange
        return out

    def _send(self, msg, exchange=None, routing_key=None, **kwargs):
        r"""Send a message.

        Args:
            msg (str, bytes): Message to be sent.
            exchange (str, optional): Exchange that message should be sent
                to. Defaults to self.exchange.
            routing_key (str, optional): Key that exchange should use to route
                the message. Defaults to self.queue.
            **kwargs: Additional keyword arguments are passed to
                :method:`pika.BlockingChannel.basic_publish`.

        Returns:
            bool: Success or failure of send.

        """
        if exchange is None:
            exchange = self.exchange
        if routing_key is None:
            routing_key = self.queue
        kwargs.setdefault('mandatory', True)
        with self.rmq_lock:
            try:
                self.channel.basic_publish(exchange, routing_key, msg, **kwargs)
            except pika.exceptions.UnroutableError:  # pragma: debug
                return False
        return True

    def _recv(self):
        r"""Receive a message.

        Returns:
            tuple (bool, obj): Success or failure of receive and received
                message.

        """
        with self.rmq_lock:
            method_frame, props, msg = self.channel.basic_get(
                queue=self.queue, auto_ack=False)
            if method_frame:
                self.channel.basic_ack(method_frame.delivery_tag)
            else:  # pragma: debug
                raise NoMessages("No messages in connection.")
        return (True, msg)

    def purge(self):
        r"""Remove all messages from the associated queue."""
        if not self._closing.has_started():
            with self.rmq_lock:
                with self._closing_thread.lock:
                    if self.is_open:
                        self.channel.queue_purge(queue=self.queue)
        super(RMQComm, self).purge()



from yggdrasil.drivers.InterpretedModelDriver import InterpretedModelDriver


class DummyModelDriver(InterpretedModelDriver):
    r"""Class that stands-in to act as a model utilizing unmatched
    input/output channels."""
    
    executable_type = 'other'
    language = 'dummy'
    full_language = False
    base_languages = ['python']
    language_ext = []
    no_executable = True
    comms_implicit = True

    def __init__(self, *args, **kwargs):
        self.runner = kwargs['runner']
        super(DummyModelDriver, self).__init__(*args, **kwargs)

    @classmethod
    def is_language_installed(self):
        r"""Determine if this model driver is installed on the current
        machine.

        Returns:
            bool: Truth of if this model driver can be run on the current
                machine.

        """
        # This is being run so python exists
        return True

    @classmethod
    def is_configured(cls):
        r"""Determine if the appropriate configuration has been performed (e.g.
        installation of supporting libraries etc.)

        Returns:
            bool: True if the language has been configured.

        """
        # There are not any config options
        return True

    @classmethod
    def language_version(cls, **kwargs):
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to cls.run_executable.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        return '0'
    
    def before_start(self):
        r"""Actions to perform before the run starts."""
        pass
                                
    def before_loop(self):
        r"""Actions before loop."""
        pass

    def run_loop(self):
        r"""Loop to check if model is still running and forward output."""
        for drv in self.runner.modeldrivers.values():
            if (drv['name'] != self.name) and drv['instance'].is_alive():
                self.wait_flag_attr('break_flag', timeout=1.0)
                return
        self.set_break_flag()

    @property
    def service_partner(self):
        r"""dict: YAML representation of the dummy model that should
        stand-in for the model client-side."""
        out = {'inputs': [], 'outputs': []}
        dir2opp = {'input': 'output', 'output': 'input'}
        for io1, io2 in dir2opp.items():
            for drv in self.yml['%s_drivers' % io1]:
                name = drv[io1 + 's'][0]['name']
                comm = getattr(drv['instance'], '%scomm' % io2[0])
                x = comm.opp_comm_kwargs(for_yaml=True)
                x['name'] = name.split(
                    drv[io1 + 's'][0]['partner_model'] + ':')[-1]
                out[io2 + 's'].append(x)
                if drv['instance']._connection_type.startswith('rpc_'):
                    assert(io2 == 'input')
                    out[io1 + 's'].append({'name': x['name'] + '_response'})
                    out['is_server'] = {io2: x['name'],
                                        io1: x['name'] + '_response'}
        for io1, io2 in dir2opp.items():
            assert(out[io2 + 's'])
            # TODO: Is there a case where a DummyModelDriver will be created
            # for a model that does not have inputs or outputs?
            # if not out[io2 + 's']:
            #     out.pop(io2 + 's')
        return out



import pytest
from yggdrasil.config import ygg_cfg
from yggdrasil.communication import new_comm
from yggdrasil.communication.RMQComm import RMQComm, check_rmq_server


_rmq_installed = RMQComm.is_installed(language='python')


@pytest.mark.skipif(not _rmq_installed, reason="RMQ Server not running")
def test_running():
    r"""Test checking for server w/ URL."""
    # url = "amqp://guest:guest@localhost:5672/%2F"
    vhost = ygg_cfg.get('rmq', 'vhost', '')
    if not vhost:
        vhost = '%2F'
    url = "amqp://%s:%s@%s:%s/%s" % (
        ygg_cfg.get('rmq', 'user', 'guest'),
        ygg_cfg.get('rmq', 'password', 'guest'),
        ygg_cfg.get('rmq', 'host', 'localhost'),
        ygg_cfg.get('rmq', 'port', '5672'),
        vhost)
    assert(check_rmq_server(url=url))
    

@pytest.mark.skipif(_rmq_installed, reason="RMQ Server running")
def test_not_running():
    r"""Test raise of an error if a RMQ server is not running."""
    comm_kwargs = dict(commtype='rmq', direction='send',
                       reverse_names=True)
    with pytest.raises(RuntimeError):
        new_comm('test', **comm_kwargs)



import pytest
from tests.metaschema.properties.test_MetaschemaProperty import (
    TestMetaschemaProperty as base_class)
import numpy as np


class TestTypeMetaschemaProperty(base_class):
    r"""Test class for TypeMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.TypeMetaschemaProperty'
    _cls = 'TypeMetaschemaProperty'
    
    @pytest.fixture(scope="class")
    def valid(self):
        r"""Objects that are valid."""
        return [(np.int8(1), 'int'), (np.int8(1), 'scalar')]

    @pytest.fixture(scope="class")
    def invalid(self):
        r"""Objects that are invalid."""
        return [(np.int8(1), 'float'), (np.float32(1), 'int')]

    @pytest.fixture(scope="class")
    def encode_errors(self):
        r"""Object that enduce errors during encoding."""
        return [np]  # Can't encode modules

    @pytest.fixture(scope="class")
    def valid_compare(self):
        r"""Objects that successfully compare."""
        return [('int', 'int'), ('int', 'scalar'), ('ply', 'object')]

    @pytest.fixture(scope="class")
    def invalid_compare(self):
        r"""Objects that do not successfully compare."""
        return [('int', 'float'), ('array', 'object'),
                ('ply', 'array'), ('1darray', 'scalar')]



import pytest
import copy
from tests import TestComponentBase as base_class
import os


class TestDriver(base_class):
    r"""Test parameters for basic Driver test class."""

    @pytest.fixture(scope="class")
    def working_dir(self):
        r"""Working director."""
        return os.path.dirname(__file__)

    @pytest.fixture
    def namespace(self, uuid):
        r"""Unique name for the test communicators."""
        return f"TESTING_{uuid}"

    @pytest.fixture
    def name(self, class_name, uuid):
        r"""str: Name of the test driver."""
        return f'Test{class_name}_{uuid}'

    @pytest.fixture
    def instance_args(self, name):
        r"""Arguments for a new instance of the tested class."""
        return tuple([name])

    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, working_dir,
                        polling_interval, namespace):
        r"""Keyword arguments for a new instance of the tested class."""
        yml = {}
        if working_dir:
            yml['working_dir'] = working_dir
        return dict(copy.deepcopy(testing_options.get('kwargs', {})), yml=yml,
                    timeout=timeout, sleeptime=polling_interval,
                    namespace=namespace)

    @pytest.fixture
    def instance(self, python_class, instance_args, instance_kwargs,
                 verify_count_threads, verify_count_comms, verify_count_fds,
                 working_dir, is_installed):
        r"""New instance of the python class for testing."""
        curpath = os.getcwd()
        if working_dir:
            os.chdir(working_dir)
        out = python_class(*instance_args, **instance_kwargs)
        try:
            yield out
        finally:
            if not out.was_started:
                out.cleanup()
            out.disconnect()
            del out
            os.chdir(curpath)

    @pytest.fixture(scope="class")
    def before_instance_started(self):
        r"""Actions performed after teh instance is created, but before it
        is started."""
        def before_instance_started_w(x):
            pass
        return before_instance_started_w

    @pytest.fixture(scope="class")
    def after_instance_started(self):
        r"""Action taken after the instance is started, but before tests
        begin."""
        def after_instance_started_w(x):
            pass
        return after_instance_started_w
        
    @pytest.fixture
    def started_instance(self, instance, before_instance_started,
                         after_instance_started):
        r"""Started version of the instance."""
        before_instance_started(instance)
        instance.start()
        try:
            after_instance_started(instance)
            yield instance
        finally:
            if not instance.was_terminated:
                instance.terminate()
            if instance.is_alive():
                instance.join()
            instance.cleanup()
            assert(not instance.is_alive())

    @pytest.fixture(scope="class")
    def assert_before_stop(self):
        r"""Assertions to make before stopping the driver instance."""
        def assert_before_stop_w():
            pass
        return assert_before_stop_w

    @pytest.fixture(scope="class")
    def run_before_terminate(self):
        r"""Commands to run while the instance is running, before terminate."""
        def run_before_terminate_w():
            pass
        return run_before_terminate_w

    @pytest.fixture(scope="class")
    def run_before_stop(self):
        r"""Commands to run while the instance is running."""
        def run_before_stop_w():
            pass
        return run_before_stop_w

    @pytest.fixture
    def assert_after_terminate(self, started_instance):
        r"""Assertions to make after terminating the driver instance."""
        def assert_after_terminate_w():
            assert(not started_instance.is_alive())
        return assert_after_terminate_w

    @pytest.fixture
    def assert_after_stop(self, assert_after_terminate):
        r"""Assertions to make after stopping the driver instance."""
        def assert_after_stop_w():
            assert_after_terminate()
        return assert_after_stop_w

    def test_init_del(self, started_instance):
        r"""Test driver creation and deletion."""
        started_instance.printStatus()
        started_instance.printStatus(return_str=True)

    def test_run_stop(self, started_instance, assert_before_stop,
                      run_before_stop, assert_after_stop, polling_interval):
        r"""Start the thread, then stop it."""
        assert_before_stop()
        run_before_stop()
        started_instance.wait(polling_interval)
        started_instance.stop()
        started_instance.stop()
        assert_after_stop()

    def test_run_terminate(self, started_instance, assert_before_stop,
                           run_before_terminate, assert_after_terminate):
        r"""Start the thread, then terminate it."""
        assert_before_stop()
        run_before_terminate()
        started_instance.terminate()
        # Second time to ensure it is escaped
        started_instance.terminate()
        if started_instance.is_alive():  # pragma: debug
            started_instance.join()
        assert_after_terminate()



import os
from yggdrasil import tools
from yggdrasil.runner import run
tyk_dir = 'tyk4'

try:
    # Change to the solution directory
    old_dir = os.getcwd()
    if not old_dir.endswith(tyk_dir):
        os.chdir(os.path.join('solutions', tyk_dir))
    if not os.path.isdir('output'):
        os.mkdir('output')

    # Part 1: shoot calling light and weather
    tools.display_source_diff('../../models/shoot_v2.py',
                              'models/shoot_calling_light_and_weather.py',
                              number_lines=True)
    tools.display_source('yamls/shoot_calling_light_and_weather.yml', number_lines=True)
    run(['yamls/shoot_calling_light_and_weather.yml'], production_run=True)

    # Part 2: light-calling-weather
    tools.display_source_diff('../../models/light_v0.py',
                              'models/light_calling_weather.py', number_lines=True)
    tools.display_source('yamls/light_calling_weather.yml', number_lines=True)
    run(['yamls/light_calling_weather.yml'], production_run=True)

    # Part 3: shoot calling light calling weather
    tools.display_source_diff('../../models/shoot_v2.py',
                              'models/shoot_calling_light_calling_weather.py',
                              number_lines=True)
    tools.display_source('yamls/shoot_calling_light_calling_weather.yml',
                         number_lines=True)
    run(['yamls/shoot_calling_light_calling_weather.yml'],
        production_run=True)

finally:
    os.chdir(old_dir)



import pytest
import os
import copy
from yggdrasil import constants
from yggdrasil.languages import get_language_ext
from yggdrasil.examples import _example_dir
from yggdrasil.metaschema.datatypes import encode_type
from yggdrasil.components import import_component
from tests.examples import TestExample as base_class


_untyped_lang = sorted([k for k, v in constants.LANGUAGE_PROPERTIES.items()
                        if (v['full_language'] and not v['is_typed'])])


class TestExampleTransforms(base_class):
    r"""Test the Transforms example."""

    parametrize_language = _untyped_lang
    parametrize_transform = ['table', 'array', 'pandas']

    @pytest.fixture(scope="class")
    def example_name(self):
        r"""str: Name of example being tested."""
        return "transforms"

    @pytest.fixture(scope="class", autouse=True)
    def transform(self, request):
        r"""str: Transform being tested."""
        return request.param

    @pytest.fixture(scope="class")
    def env(self, example_name, language, transform, example_module):
        r"""dict: Environment variables set for the test."""
        language_ext = get_language_ext(language)
        modelfile = os.path.join(_example_dir, example_name,
                                 'src', 'model' + language_ext)
        drv = import_component('model', language)
        testdata = example_module.get_test_data(transform)
        testtype = encode_type(testdata)
        inputs = [{'name': 'x',
                   'datatype': copy.deepcopy(testtype)}]
        outputs = [{'name': 'y',
                    'datatype': copy.deepcopy(testtype)}]
        # Write the model
        function_contents = []
        for i, o in zip(inputs, outputs):
            function_contents += drv.write_assign_to_output(
                o, i, outputs_in_inputs=drv.outputs_in_inputs)
        lines = drv.write_function_def(
            'model', function_contents=function_contents,
            inputs=copy.deepcopy(inputs),
            outputs=copy.deepcopy(outputs),
            outputs_in_inputs=drv.outputs_in_inputs,
            opening_msg='IN MODEL', closing_msg='MODEL EXIT',
            print_inputs=True, print_outputs=True)
        with open(modelfile, 'w') as fd:
            print(modelfile)
            print('\n'.join(lines))
            fd.write('\n'.join(lines))
        env = {}
        env['TEST_LANGUAGE'] = language
        env['TEST_LANGUAGE_EXT'] = language_ext
        env['TEST_TRANSFORM'] = transform
        if transform == 'table':
            env['TEST_MODEL_IO'] = (
                'outputs:\n'
                + '      - name: '
                + language + '_model:output\n'
                + '        format_str: "%s\\t%d\\t%f\\n"')
        try:
            yield env
        finally:
            if os.path.isfile(modelfile):
                os.remove(modelfile)

    @pytest.fixture
    def check_results(self):
        r"""This should be overridden with checks for the result."""
        def check_results_w():
            pass
        return check_results_w



import shutil
from yggdrasil.components import import_component
from yggdrasil.drivers.ModelDriver import ModelDriver
from yggdrasil.multitasking import MPI, MPIRequestWrapper


class MPIPartnerModel(ModelDriver):
    r"""Class for shadowing a model run on another MPI process."""

    _schema_subtype_description = ('Model is being run on another MPI '
                                   'process and this driver is used as '
                                   'as stand-in to monitor it on the root '
                                   'process.')
    executable_type = 'other'
    language = 'mpi'
    full_language = False
    base_languages = []
    language_ext = []
    comms_implicit = True

    def __init__(self, *args, **kwargs):
        kwargs.pop('function', None)
        super(MPIPartnerModel, self).__init__(*args, **kwargs)
        self.partner_driver = import_component('model',
                                               self.yml['partner_driver'],
                                               without_schema=True)
        self.partner_driver.mpi_partner_init(self)

    def cleanup(self, *args, **kwargs):
        r"""Remove compile executable."""
        self.partner_driver.mpi_partner_cleanup(self)
        super(MPIPartnerModel, self).cleanup(*args, **kwargs)

    @classmethod
    def is_language_installed(self):
        r"""Determine if this model driver is installed on the current
        machine.

        Returns:
            bool: Truth of if this model driver can be run on the current
                machine.

        """
        return (MPI is not None)

    @classmethod
    def is_configured(cls):
        r"""Determine if the appropriate configuration has been performed (e.g.
        installation of supporting libraries etc.)

        Returns:
            bool: True if the language has been configured.

        """
        return True

    @classmethod
    def is_comm_installed(cls, **kwargs):
        r"""Determine if a comm is installed for the associated programming
        language.

        Args:
            **kwargs: Keyword arguments are ignored.

        Returns:
            bool: True if a comm is installed for this language.

        """
        return True
    
    @classmethod
    def language_version(cls, **kwargs):
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to cls.run_executable.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        import mpi4py
        return 'mpi4py %s' % mpi4py.__version__
    
    @classmethod
    def language_executable(cls, **kwargs):
        r"""Command required to compile/run a model written in this language
        from the command line.

        Returns:
            str: Name of (or path to) compiler/interpreter executable required
                to run the compiler/interpreter from the command line.

        """
        return shutil.which('mpiexec')
        
    def run_model(self, **kwargs):
        r"""Dummy stand-in for ModelDriver run_model method."""
        return None
        
    def before_start(self, **kwargs):
        r"""Actions to perform before the run starts."""
        kwargs['no_queue_thread'] = True
        super(MPIPartnerModel, self).before_start(**kwargs)

    # def init_mpi_env(self):
    #     r"""Send env information to the partner model."""
    #     env = copy.deepcopy(self.env)
    #     env.update(self.get_io_env())
    #     self.send_mpi(env, tag=self._mpi_tags['ENV'])
        
    def init_mpi(self):
        r"""Initialize MPI communicator."""
        self.send_mpi('START', tag=self._mpi_tags['START'])
        self._mpi_requests['stopped'] = MPIRequestWrapper(
            self.recv_mpi(tag=self._mpi_tags['STOP_RANK0'], dont_block=True))

    def stop_mpi_partner(self, **kwargs):
        r"""Send a message to stop the MPI partner model on the main process."""
        kwargs.update(dest=self._mpi_partner_rank,
                      tag=self._mpi_tags['STOP_RANKX'],
                      msg=self.n_sent_messages)
        super(MPIPartnerModel, self).stop_mpi_partner(**kwargs)

    def run_loop(self):
        r"""Loop to check if model is still running."""
        if self.check_mpi_request('stopped'):
            self.set_break_flag()
        else:
            self.sleep()

    def graceful_stop(self):
        r"""Gracefully stop the driver."""
        if self.has_sent_messages:
            self.wait_on_mpi_request('stopped', timeout=10)
        self.set_break_flag()
        super(MPIPartnerModel, self).graceful_stop()
        
    def kill_process(self):
        r"""Kill the process running the model, checking return code."""
        self.set_break_flag()
        if not self.model_process_complete:  # pragma: debug
            self._mpi_requests['stopped'].completed = True
        super(MPIPartnerModel, self).kill_process()
        
    @property
    def model_process_complete(self):
        r"""bool: Has the process finished or not. Returns True if the process
        has not started."""
        return self.check_mpi_request('stopped')



import pytest
from yggdrasil.metaschema.datatypes import conversions


def test_register_conversion_errors():
    r"""Test errors in register_conversion."""
    x = list(conversions._conversion_registry.keys())[0]
    with pytest.raises(ValueError):
        conversions.register_conversion(x[0], x[1])


def test_ply2obj(ply_test_value, obj_test_value):
    r"""Test conversion between ply and obj objects."""
    from yggdrasil.metaschema.datatypes import (
        ObjMetaschemaType, PlyMetaschemaType)
    obj_class = ObjMetaschemaType.ObjMetaschemaType
    ply_class = PlyMetaschemaType.PlyMetaschemaType
    # Start at ply
    obj = conversions.ply2obj(ply_test_value)
    obj_class.validate_instance(obj, {'type': 'obj'})
    ply = conversions.obj2ply(obj)
    ply_class.validate_instance(ply, {'type': 'ply'})
    # Start at obj
    ply = conversions.obj2ply(obj_test_value)
    ply_class.validate_instance(ply, {'type': 'ply'})
    obj = conversions.ply2obj(ply)
    obj_class.validate_instance(obj, {'type': 'obj'})



import pytest
import os
import glob
import tempfile
import shutil
import importlib
from yggdrasil.components import import_component
from yggdrasil import runner, tools, platform
from yggdrasil.examples import (
    get_example_yaml, get_example_source, get_example_languages,
    ext_map, display_example, source)
from tests import TestBase as base_class


_ext2lang = {v: k for k, v in ext_map.items()}
_test_registry = {}
_default_comm = tools.get_default_comm()


@pytest.mark.suite("examples", disabled=True)
def test_get_example_yaml():
    r"""Test get_example_yaml."""
    with pytest.raises(KeyError):
        get_example_yaml('invalid', 'invalid')
    with pytest.raises(KeyError):
        get_example_yaml('hello', 'invalid')
    get_example_yaml('hello', 'r')
    get_example_yaml('hello', 'R')


@pytest.mark.suite("examples", disabled=True)
def test_get_example_source():
    r"""Test get_example_source."""
    with pytest.raises(KeyError):
        get_example_source('invalid', 'invalid')
    with pytest.raises(KeyError):
        get_example_source('hello', 'invalid')
    get_example_source('hello', 'r')
    get_example_source('hello', 'R')


@pytest.mark.suite("examples", disabled=True)
def test_get_example_languages():
    r"""Test get_example_languages."""
    with pytest.raises(KeyError):
        get_example_languages('invalid')
    get_example_languages('ascii_io')
    get_example_languages('ascii_io', language='python')
    get_example_languages('ascii_io', language='all')
    get_example_languages('ascii_io', language='all_nomatlab')


@pytest.mark.suite("examples", disabled=True)
def test_display_example():
    r"""Test display_example."""
    display_example('hello', 'r')


_examples = sorted([x for x in source.keys() if x not in
                    ['SaM', 'ascii_io', 'conditional_io', 'rpcFib',
                     'rpc_lesson1', 'rpc_lesson2', 'rpc_lesson2b',
                     'timed_pipe', 'transforms', 'types',
                     # Below have explicit tests so they can be run with MPI
                     'gs_lesson4', 'rpc_lesson3b', 'model_error_with_io']])


@pytest.mark.suite("examples", disabled=True)
class TestExample(base_class):
    r"""Base class for running examples."""

    parametrize_example_name = _examples

    @pytest.fixture(scope="class")
    def example_name(self, request):
        r"""str: Name of example being tested."""
        return request.param

    @pytest.fixture(scope="class", autouse=True)
    def language(self, request, example_name, check_required_languages):
        r"""str: Language of the currect test."""
        avail_langs = get_example_languages(example_name)
        lang = tools.is_language_alias(request.param, avail_langs)
        if not lang:
            pytest.skip(f"example dosn't have a {request.param} version "
                        f"(supported: {avail_langs})")
        check_required_languages(
            get_example_languages(example_name, language=lang))
        return lang

    @pytest.fixture(scope="class", autouse=True)
    def prepare_makefile(self, language, mpi_rank, yamldir):
        r"""Prepare a make file for the test."""
        if language == 'make':
            makefile = os.path.join(yamldir, 'src', 'Makefile')
            if platform._is_win:  # pragma: windows
                makedrv = import_component('model', 'make')
                assert(makedrv.get_tool('compiler').toolname == 'nmake')
                make_ext = '_windows'
            else:
                make_ext = '_linux'
            if not os.path.isfile(makefile):
                shutil.copy(makefile + make_ext, makefile)
            try:
                yield
            finally:
                if (mpi_rank == 0) and os.path.isfile(makefile):
                    os.remove(makefile)
        else:
            yield

    @pytest.fixture(scope="class")
    def example_module(self, example_name):
        r"""Python module associated with the test."""
        try:
            return importlib.import_module(
                f'yggdrasil.examples.{example_name}')
        except ImportError:
            # pytest.skip(f"Could not import {example_name} example module")
            return None

    @pytest.fixture(scope="class")
    def testing_options(self, example_module):
        r"""dict: Testing options."""
        if hasattr(example_module, 'get_testing_options'):
            return example_module.get_testing_options()
        return {}

    @pytest.fixture(scope="class")
    def expects_error(self, testing_options):
        r"""bool: True if the example expects to raise an error."""
        return testing_options.get('expects_error', False)

    @pytest.fixture(scope="class")
    def env(self, testing_options):
        r"""dict: Environment variables set for the test."""
        return testing_options.get('env', {})

    @pytest.fixture
    def setup_env(self, env):
        r"""Setup the env."""
        old_val = {k: os.environ.get(k, None) for k in env.keys()}
        os.environ.update(env)
        yield
        for k, v in old_val.items():
            if v is None:
                del os.environ[k]
            else:
                os.environ[k] = v

    @pytest.fixture
    def namespace(self, example_name, uuid):
        r"""str: Namespace for the example."""
        return f"{example_name}_{uuid}"

    @pytest.fixture(scope="class")
    def tempdir(self):
        r"""str: Temporary directory."""
        return tempfile.gettempdir()

    @pytest.fixture(scope="class")
    def yaml(self, example_name, language):
        r"""str: The full path to the yaml file for this example."""
        return get_example_yaml(example_name, language)

    @pytest.fixture(scope="class")
    def yamldir(self, yaml):
        r"""str: Full path to the directory containing the yaml file."""
        if yaml is None:  # pragma: no cover
            return None
        if isinstance(yaml, list):
            out = os.path.dirname(yaml[0])
        else:
            out = os.path.dirname(yaml)
        return out

    @pytest.fixture(scope="class")
    def input_dir(self, testing_options, yamldir):
        r"""str: Directory containing input files."""
        out = testing_options.get('input_dir', 'Input')
        if out:
            return os.path.join(yamldir, out)
        return yamldir

    @pytest.fixture(scope="class")
    def output_dir(self, testing_options, yamldir, tempdir):
        r"""str: Directory containing output files."""
        if testing_options.get('temp_output_dir', False):
            return tempdir
        out = testing_options.get('output_dir', 'Output')
        if out:
            return os.path.join(yamldir, out)
        return yamldir

    @pytest.fixture(scope="class", autouse=True)
    def create_output_dir(self, output_dir):
        if not os.path.isdir(output_dir):
            os.mkdir(output_dir)

    @pytest.fixture(scope="class")
    def input_files(self, testing_options, input_dir):
        r"""list: Input files for the run."""
        out = testing_options.get('input_files', None)
        if isinstance(out, str):
            out = [out]
        if out and input_dir:
            out = [os.path.join(input_dir, x) for x in out]
        return out

    @pytest.fixture(scope="class")
    def expected_output_files(self, testing_options,
                              yamldir, input_files):  # pragma: debug
        r"""list: Examples of expected output for the run."""
        if 'expected_output_files' in testing_options:
            out = testing_options['expected_output_files']
            if isinstance(out, str):
                out = [out]
            out = [os.path.join(yamldir, x) for x in out]
            return out
        return input_files

    @pytest.fixture(scope="class")
    def output_files(self, testing_options, output_dir):
        r"""list: Output files for the run."""
        out = testing_options.get('output_files', None)
        if isinstance(out, str):
            out = [out]
        if out and output_dir:
            out = [os.path.join(output_dir, x) for x in out]
        return out

    @pytest.fixture(scope="class")
    def core_dump(self, yamldir):
        r"""str: Pattern for core dump that may be produced."""
        return os.path.join(yamldir, 'core.*')

    @pytest.fixture(scope="class")
    def read_file(self):
        r"""Read a file."""
        def read_file_w(fname):
            with open(fname, 'r') as fd:
                return fd.read()
        return read_file_w

    @pytest.fixture
    def results(self, expected_output_files, read_file):
        r"""list: Results that should be found in the output files."""
        if expected_output_files is None:  # pragma: debug
            return None
        out = []
        for fname in expected_output_files:
            assert(os.path.isfile(fname))
            out.append(read_file(fname))
        return out

    @pytest.fixture
    def check_results(self, results, output_files, check_file_exists,
                      check_file_size, check_file_contents, testing_options):
        r"""This should be overridden with checks for the result."""
        def check_results_w():
            if testing_options.get('validation_function', False):
                testing_options['validation_function']()
            if testing_options.get('skip_check_results', False):
                return
            res_list = results
            out_list = output_files
            if (out_list is None) or (res_list is None):
                return
            assert(res_list is not None)
            assert(out_list is not None)
            assert(len(res_list) == len(out_list))
            for res, fout in zip(res_list, out_list):
                check_file_exists(fout)
                if isinstance(res, tuple):
                    res[0](fout, *res[1:])
                else:
                    check_file_size(fout, res)
                    check_file_contents(fout, res)
        return check_results_w

    @pytest.fixture
    def example_cleanup(self, yaml, output_files, core_dump):
        r"""Cleanup files created during the test."""
        def example_cleanup_w():
            if (yaml is not None) and (output_files is not None):
                for fout in output_files:
                    if os.path.isfile(fout):
                        tools.remove_path(fout, timeout=5)
            for f in glob.glob(core_dump):
                os.remove(f)
        return example_cleanup_w
    
    def test_example(self, example_name, language, namespace, yaml,
                     setup_env, expects_error, on_mpi,
                     mpi_rank, check_results, example_cleanup,
                     adv_global_mpi_tag, optionally_disable_verify_count_fds):
        r"""This runs an example in the correct language."""
        if example_name.startswith('timesync'):
            # Timesync examples include ploting in the verification script
            # which opens file descriptors that are not promptly cleaned up
            optionally_disable_verify_count_fds()
        assert(yaml is not None)
        # Run
        mpi_tag_start = None
        if on_mpi:
            mpi_tag_start = adv_global_mpi_tag(1000)
        r = runner.get_runner(yaml, namespace=namespace,
                              production_run=True,
                              mpi_tag_start=mpi_tag_start)
        try:
            try:
                r.run()
            except runner.IntegrationError:
                pass
            r.printStatus()
            r.printStatus(return_str=True)
            if mpi_rank != 0:
                return
            if expects_error:
                assert(r.error_flag)
            else:
                assert(not r.error_flag)
            try:
                check_results()
            except BaseException:  # pragma: debug
                if r is not None:
                    r.printStatus()
                raise
        finally:
            if mpi_rank == 0:
                example_cleanup()



#!/usr/bin/python
import os
import sys
import copy
import logging
import subprocess
import argparse
import pprint
import shutil
from yggdrasil import constants
LANGUAGES = getattr(constants, 'LANGUAGES', {})
LANGUAGES_WITH_ALIASES = getattr(constants, 'LANGUAGES_WITH_ALIASES', {})


logger = logging.getLogger(__name__)
package_dir = os.path.dirname(os.path.abspath(__file__))


def githook():
    r"""Git hook to determine if the Github workflow need to be
    re-generated."""
    try:
        files = subprocess.check_output(
            ["git", "diff-index", "--cached", "--name-only",
             "--diff-filter=ACMRTUXB", "HEAD"],
            stderr=subprocess.PIPE).decode('utf-8').splitlines()
    except subprocess.CalledProcessError:
        return 1
    regen = (os.path.join('utils', 'test-install-base.yml') in files)
    if regen:
        try:
            gitdir = subprocess.check_output(
                ["git", "rev-parse", "--git-dir"],
                stderr=subprocess.PIPE).decode('utf-8').strip()
        except subprocess.CalledProcessError:
            return 1
        workflow_dir = os.path.join(gitdir, '..', '.github', 'workflows')
        generate_gha_workflow(args=[], gitdir=gitdir)
        try:
            subprocess.run(
                ['git', 'add',
                 os.path.join(workflow_dir, 'test-install.yml')],
                check=True,
                stderr=subprocess.STDOUT,
                stdout=subprocess.PIPE)
        except subprocess.CalledProcessError:
            return 1
    return 0


class ArgumentTuple(tuple):

    def __new__(self, args, kwargs):
        return tuple.__new__(ArgumentTuple, (args, kwargs))


class ConditionalArgumentTuple(ArgumentTuple):

    def __new__(cls, args, kwargs, conditions=None):
        out = ArgumentTuple.__new__(ConditionalArgumentTuple,
                                    args, kwargs)
        out.conditions = conditions
        if out.conditions is None:
            out.conditions = {}
        return out
            

class ArgumentBase(object):
    
    def __init__(self, arguments=None, conditions=None, **kwargs):
        self.arguments = arguments
        if self.arguments is None:
            self.arguments = []
        self.conditions = conditions
        if self.conditions is None:
            self.conditions = {}
        self.kwargs = kwargs


class ArgumentParser(ArgumentBase):
    pass


class ArgumentGroup(ArgumentBase):

    def __init__(self, exclusive=False, **kwargs):
        self.exclusive = exclusive
        super(ArgumentGroup, self).__init__(**kwargs)


class ArgumentSubparser(ArgumentBase):

    def __init__(self, parsers=None, **kwargs):
        self.parsers = parsers
        if self.parsers is None:
            self.parsers = []
        super(ArgumentSubparser, self).__init__(**kwargs)


class SubCommandMeta(type):
    r"""Meta class for subcommands."""

    def __call__(cls, *args, **kwargs):
        return cls.call(*args, **kwargs)


def ReplacementWarning(old, new):
    import warnings
    warnings.warn(("'%s' will soon be removed. Use '%s' instead.")
                  % (old, new), FutureWarning)


class SubCommand(metaclass=SubCommandMeta):
    r"""Class for handling subcommands so that they can be run
    as subcommands or individually."""

    name = None
    help = None
    arguments = []
    allow_unknown = False

    @classmethod
    def parse_args(cls, parser, args=None, allow_unknown=False,
                   namespace=None):
        # TODO: Check choices for positional arguments that can
        # have more than one element
        if isinstance(args, argparse.Namespace):
            return args
        kws = dict(args=args)
        if namespace is not None:
            kws['namespace'] = namespace
        if cls.allow_unknown or allow_unknown:
            args, extra = parser.parse_known_args(**kws)
            args._extra_commands = extra
        else:
            args = parser.parse_args(**kws)
        for k in ['language', 'languages']:
            v = getattr(args, k, None)
            if isinstance(v, list):
                v_flag = getattr(args, k + '_flag', None)
                if isinstance(v_flag, list):
                    v.extend(v_flag)
                if (len(v) == 0) or ('all' in v):
                    setattr(args, k, LANGUAGES.get('all', []))
                    args.all_languages = True
        return args

    @classmethod
    def func(cls, args):  # pragma: debug
        raise NotImplementedError

    @classmethod
    def call(cls, args=None, namespace=None, **kwargs):
        parser = cls.get_parser(args=args)
        args = cls.parse_args(parser, args=args, namespace=namespace)
        return cls.func(args, **kwargs)

    @classmethod
    def get_parser(cls, args=None):
        parser = argparse.ArgumentParser(cls.help)
        cls.add_arguments(parser, args=args)
        return parser

    @classmethod
    def add_argument_to_parser(cls, parser, x):
        if hasattr(x, 'conditions') and x.conditions:
            for k, v in x.conditions.items():
                if k == 'os':
                    from yggdrasil import platform
                    if platform._platform not in v:
                        return
                else:  # pragma: debug
                    raise NotImplementedError(k)
        if isinstance(x, list):
            for xx in x:
                cls.add_argument_to_parser(parser, xx)
        elif isinstance(x, (tuple, ArgumentTuple,
                            ConditionalArgumentTuple)):
            assert(len(x) == 2)
            args, kwargs = x[:]
            try:
                parser.add_argument(*args, **kwargs)
            except ValueError:
                if kwargs.get('action', None) == 'extend':
                    kwargs['action'] = 'append'
                    kwargs.pop('nargs', None)
                    parser.add_argument(*args, **kwargs)
                else:
                    raise
        elif isinstance(x, ArgumentGroup):
            if x.exclusive:
                group = parser.add_mutually_exclusive_group(**x.kwargs)
            else:
                group = parser.add_argument_group(**x.kwargs)
            cls.add_argument_to_parser(group, x.arguments)
        elif isinstance(x, ArgumentSubparser):
            subparsers = parser.add_subparsers(**x.kwargs)
            for xx in x.parsers:
                isubparser = subparsers.add_parser(**xx.kwargs)
                cls.add_argument_to_parser(isubparser, x.arguments)
                cls.add_argument_to_parser(isubparser, xx.arguments)
        else:
            raise NotImplementedError("type(x) = %s" % type(x))

    @classmethod
    def add_arguments(cls, parser, args=None):
        cls.add_argument_to_parser(parser, cls.arguments)

    @classmethod
    def add_subparser(cls, subparsers, args=None):
        parser = subparsers.add_parser(cls.name, help=cls.help)
        cls.add_arguments(parser, args=args)
        parser.set_defaults(func=cls.func)


class yggrun(SubCommand):
    r"""Start a run."""

    name = "run"
    help = "Run an integration."
    arguments = [
        (('yamlfile', ),
         {'nargs': '+',
          'help': "One or more yaml specification files."}),
        (('--with-mpi', '--mpi-nproc'),
         {'type': int, 'default': 1,
          'help': 'Number of MPI processes to run on.'}),
        (('--mpi-tag-start', ),
         {'type': int, 'default': 0,
          'help': 'Tag that MPI communications should start at.'}),
        (('--validate', ),
         {'action': 'store_true',
          'help': ('Validate the run via model validation commands on '
                   'completion.')})]

    @classmethod
    def add_arguments(cls, parser, **kwargs):
        from yggdrasil import config
        super(yggrun, cls).add_arguments(parser, **kwargs)
        config.get_config_parser(parser, skip_sections='testing')

    @classmethod
    def func(cls, args):
        if args.with_mpi > 1:
            new_args = ['mpiexec', '-n', str(args.with_mpi)]
            i = 0
            while i < len(sys.argv):
                x = sys.argv[i]
                if x.startswith(('--with-mpi', '--mpi-nproc')):
                    if '=' not in x:
                        i += 1
                else:
                    new_args.append(x)
                i += 1
            return subprocess.check_call(new_args)
        from yggdrasil import runner, config
        prog = sys.argv[0].split(os.path.sep)[-1]
        with config.parser_config(args):
            runner.run(args.yamlfile, ygg_debug_prefix=prog,
                       production_run=args.production_run,
                       mpi_tag_start=args.mpi_tag_start,
                       validate=args.validate)


class integration_service_manager(SubCommand):
    r"""Start or manage the yggdrasil service manager."""

    name = "integration-service-manager"
    help = "Start or manage a integration service manager."
    arguments = [
        (('--manager-name', ),
         {'help': "Name that will be used to identify the service manager."}),
        (('--service-type', ),
         {'choices': ['flask', 'rmq'],
          'help': ("Type of service that should be started. If not "
                   "provided, the default will be the (\'services\', "
                   "\'default_type\') configuration option, if set, and "
                   "will otherwise be \'flask\'.")}),
        (('--commtype', ),
         {'type': str,
          'help': ("Type of communicator that should be used for connections "
                   "to services. If not provided, the default will be "
                   "determined by the (\'services\', \'default_comm\') "
                   "configuration option, if set.")}),
        (('--address', ),
         {'type': str,
          'help': ('URL for requests to the service manager. '
                   'If not provided, the default will be determined by the '
                   '(\'services\', \'default_type\') configuration option, '
                   'if set, and based on the selected \'service-type\', if '
                   'not. For a service-type of \'flask\', this should be the '
                   'http address with the port that should be used '
                   'and the default will be \'http://localhost:{port}\'. '
                   'For a service-type of \'rmq\', this should be an '
                   'amqp broker address and the default will be '
                   '\'amqp://guest:guest@localhost:{port}/%%2f\'.')}),
        (('--port', ),
         {'type': int,
          'help': ('Port that should be used to access the app. '
                   'Defaults to 5000 for a flask service manager and '
                   '5672 for a RabbitMQ service manager if not set '
                   'via the PORT environment variable.')}),
        ArgumentSubparser(
            title='action', dest='action',
            description='Management action to take.',
            parsers=[
                ArgumentParser(
                    name='start',
                    help=('Start an integration service manager and/or '
                          'integration service.'),
                    arguments=[
                        (('--remote-url', ),
                         {'type': str,
                          'help': ('URL that will be used to access '
                                   'the service manager and running '
                                   'integrations by remote requests. '
                                   'If not provided, the environment '
                                   'variable YGGDRASIL_SERVICE_HOST_URL '
                                   'will be used if it is set. If it is '
                                   'not set, the local ``address`` will '
                                   'be used under the assumption that '
                                   'the service manager and integration '
                                   'services will only be connected to '
                                   'by local integrations.')}),
                        (('--integration-name', ),
                         {'default': None,
                          'help': ('Name of integration to start. If not '
                                   'provided, a service manager will be '
                                   'started.')}),
                        (('--integration-yamls', ),
                         {'nargs': '+',
                          'help': ('One or more YAML specification files '
                                   'defining the integration. This argument '
                                   'may be omitted if \'name\' refers to a '
                                   'registered integration.')}),
                        (('--with-coverage', ),
                         {'action': 'store_true',
                          'help': ('Enable coverage cleanup for testing.')}),
                        (('--model-repository', ),
                         {'type': str,
                          'help': ('URL for a directory in a Git repository '
                                   'containing models that should be loaded '
                                   'into the service manager registry.')}),
                        (('--log-level', ),
                         {'type': int,
                          'help': ('Level of logging that should be '
                                   'performed for the service manager '
                                   'application.')})]),
                ArgumentParser(
                    name='stop',
                    help=('Stop an integration service manager or '
                          'an integration service.'),
                    arguments=[
                        (('--integration-name', ),
                         {'default': None,
                          'help': ('Name of integration to stop. If not '
                                   'provided, the service manager will be '
                                   'stopped as will all of the running '
                                   'services.')})]),
                ArgumentParser(
                    name='status',
                    help=('Get list of available services and the status '
                          'of any running services.')),
                ArgumentParser(
                    name='register',
                    help='Register an integration with the service manager.',
                    arguments=[
                        (('integration-name', ),
                         {'type': str,
                          'help': ('The name that the integration should be '
                                   'registered under or the path to a YAML '
                                   'file containing a list of one or more '
                                   'mappings between name and YAML '
                                   'specification files for integrations '
                                   'that should be registered.')}),
                        (('integration-yamls', ),
                         {'nargs': '*',
                          'help': ('One or more YAML specification files '
                                   'defining the integration.')})]),
                ArgumentParser(
                    name='unregister',
                    help='Unregister an integration with the service manager.',
                    arguments=[
                        (('integration-name', ),
                         {'type': str,
                          'help': ('The name of the integration to remove '
                                   'from the registry or the path to a YAML '
                                   'file containing a list of one or more '
                                   'mappings between integration name and '
                                   'YAML specification files for '
                                   'integrations that should be '
                                   'unregistered.')})]),
            ])]

    @classmethod
    def func(cls, args):
        from yggdrasil.services import IntegrationServiceManager
        integration_name = getattr(args, 'integration-name',
                                   getattr(args, 'integration_name', None))
        integration_yamls0 = getattr(args, 'integration-yamls',
                                     getattr(args, 'integration_yamls', None))
        integration_yamls = []
        if integration_yamls0:
            for yml in integration_yamls0:
                if not os.path.isabs(yml):
                    yml = os.path.abspath(yml)
                integration_yamls.append(yml)
        elif integration_name and os.path.isfile(integration_name):
            if not os.path.isabs(integration_name):
                integration_name = os.path.abspath(integration_name)
        for_request = (
            (args.action in ['status', 'register', 'unregister', 'stop'])
            or (integration_name is not None))
        x = IntegrationServiceManager(name=args.manager_name,
                                      service_type=args.service_type,
                                      commtype=args.commtype,
                                      address=args.address,
                                      port=args.port,
                                      for_request=for_request)
        if args.action in ['start', None]:
            if integration_name is None:
                if not x.is_running:
                    x.start_server(
                        remote_url=getattr(args, 'remote_url', None),
                        with_coverage=getattr(args, 'with_coverage', False),
                        model_repository=getattr(args, 'model_repository',
                                                 None),
                        log_level=getattr(args, 'log_level', None))
            else:
                x.send_request(integration_name,
                               yamls=integration_yamls,
                               action='start')
        elif args.action == 'stop':
            if integration_name is None:
                x.stop_server()
            else:
                x.send_request(integration_name, action='stop')
        elif args.action == 'status':
            x.printStatus()
        elif args.action == 'register':
            x.registry.add(name=integration_name,
                           yamls=integration_yamls)
        elif args.action == 'unregister':
            x.registry.remove(name=integration_name)
        else:
            raise NotImplementedError(args.action)


class ygginfo(SubCommand):
    r"""Print information about yggdrasil installation."""

    name = 'info'
    help = ('Display information about the current yggdrasil '
            'installation.')
    arguments = [
        (('--no-languages', ),
         {'action': 'store_true', 'dest': 'no_languages',
          'help': ('Don\'t print information about individual '
                   'languages.')}),
        (('--no-comms', ),
         {'action': 'store_true', 'dest': 'no_comms',
          'help': ('Don\'t print information about individual '
                   'comms.')}),
        (('--verbose', '-v'),
         {'action': 'store_true',
          'help': ('Increase the verbosity of the printed '
                   'information.')}),
        ArgumentSubparser(
            title='tool', dest='tool',
            description='Compilation tool types to get info about.',
            arguments=[
                (('language', ),
                 {'choices': LANGUAGES_WITH_ALIASES.get('compiled', []),
                  'type': str.lower,
                  'help': 'Language to get tool information for.'}),
                (('--toolname', ),
                 {'default': None,
                  'help': ('Name of tool to get information for. '
                           'If not provided, information for the '
                           'default tool will be returned.')}),
                (('--flags', ),
                 {'action': 'store_true',
                  'help': ('Display the flags that yggdrasil will '
                           ' pass to the tool when it is called.')}),
                (('--fullpath', ),
                 {'action': 'store_true',
                  'help': 'Get the full path to the tool exectuable.'})],
            parsers=[
                ArgumentParser(
                    name='compiler',
                    help='Get information about a compiler.'),
                ArgumentParser(
                    name='linker',
                    help='Get information about a linker.',
                    arguments=[
                        (('--library', ),
                         {'action': 'store_true',
                          'help': 'Get flags for linking a library.'})]),
                ArgumentParser(
                    name='archiver',
                    help='Get information about a archiver.')])]

    @classmethod
    def func(cls, args, return_str=False):
        from yggdrasil import platform
        from yggdrasil.components import import_component
        if args.tool:
            drv = import_component('model', args.language)
            if args.flags:
                if args.tool == 'compiler':
                    flags = drv.get_compiler_flags(
                        for_model=True, toolname=args.toolname,
                        dry_run=True, dont_link=True)
                    if '/link' in flags:  # pragma: windows
                        flags = flags[:flags.index('/link')]
                    for k in ['-c']:
                        if k in flags:
                            flags.remove(k)
                else:
                    if args.tool == 'archiver':
                        libtype = 'static'
                    elif getattr(args, 'library', False):
                        libtype = 'shared'
                    else:
                        libtype = 'object'
                    flags = drv.get_linker_flags(
                        for_model=True, toolname=args.toolname,
                        dry_run=True, libtype=libtype)
                out = ' '.join(flags)
                if platform._is_win:  # pragma: windows:
                    out = out.replace('/', '-')
                    out = out.replace('\\', '/')
            elif args.fullpath:
                out = drv.get_tool(args.tool).get_executable(full_path=True)
            else:
                out = drv.get_tool(args.tool, return_prop='name')
            if return_str:
                return out
            print(out)
            return
        from yggdrasil import tools, config, __version__
        lang_list = tools.get_installed_lang()
        comm_list = tools.get_installed_comm()
        comm_lang_list = []
        prefix = '    '
        curr_prefix = ''
        vardict = [
            ('Location', os.path.dirname(__file__)),
            ('Version', __version__),
            ('Languages', ', '.join(lang_list)),
            ('Communication Mechanisms',
             ', '.join(tools.get_installed_comm())),
            ('Default Comm Mechanism', tools.get_default_comm()),
            ('Config File', config.usr_config_file)]
        try:
            # Add language information
            if not args.no_languages:
                # Install languages
                vardict.append(('Installed Languages:', ''))
                curr_prefix += prefix
                for lang in sorted(lang_list):
                    drv = import_component('model', lang)
                    vardict.append(
                        (curr_prefix + '%s:' % lang.upper(), ''))
                    if not drv.comms_implicit:
                        comm_lang_list.append(lang)
                    curr_prefix += prefix
                    exec_name = drv.language_executable()
                    if exec_name:
                        if not os.path.isabs(exec_name):
                            exec_name = shutil.which(exec_name)
                        vardict.append((curr_prefix + 'Location',
                                        exec_name))
                    vardict.append((curr_prefix + 'Version',
                                    drv.language_version()))
                    curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Not installed languages
                vardict.append(("Languages Not Installed:", ''))
                curr_prefix += prefix
                for lang in tools.get_supported_lang():
                    if lang in lang_list:
                        continue
                    drv = import_component('model', lang)
                    vardict.append(
                        (curr_prefix + '%s:' % lang.upper(), ''))
                    curr_prefix += prefix
                    vardict.append(
                        (curr_prefix + "Language Installed",
                         drv.is_language_installed()))
                    if drv.executable_type == 'compiler':
                        curr_prefix += prefix
                        vardict += [
                            (curr_prefix
                             + ("%s Installed (%s)"
                                % (x.title(),
                                   getattr(drv, 'default_%s' % x, None))),
                             drv.is_tool_installed(x))
                            for x in ['compiler', 'linker', 'archiver']]
                        curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                    vardict.append(
                        (curr_prefix + "Base Languages Installed",
                         drv.are_base_languages_installed()))
                    missing = []
                    if not drv.are_base_languages_installed(
                            missing=missing):
                        vardict.append(
                            (curr_prefix
                             + "Base Languages Not Installed",
                             missing))
                    vardict.append(
                        (curr_prefix + "Dependencies Installed",
                         drv.are_dependencies_installed()))
                    if not drv.are_dependencies_installed():
                        vardict.append(
                            (curr_prefix
                             + "Dependencies Not Installed",
                             [b for b in drv.interface_dependencies if
                              (not drv.is_library_installed(b))]))
                    vardict.append(
                        (curr_prefix + "Interface Installed",
                         drv.is_interface_installed()))
                    vardict.append((curr_prefix + "Comm Installed",
                                    drv.is_comm_installed()))
                    vardict.append((curr_prefix + "Configured",
                                    drv.is_configured()))
                    vardict.append((curr_prefix + "Disabled",
                                    drv.is_disabled()))
                    curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
            # Add comm information
            if not args.no_comms:
                # Fully installed comms
                vardict.append(
                    ('Comms Available for All Languages:', ''))
                curr_prefix += prefix
                for comm in sorted(comm_list):
                    cmm = import_component('comm', comm)
                    vardict.append(
                        (curr_prefix + '%s' % comm.upper(), ''))
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Partially installed comms
                vardict.append(
                    ('Comms Available for Some/No Languages:', ''))
                curr_prefix += prefix
                for comm in tools.get_supported_comm():
                    if comm in comm_list:
                        continue
                    cmm = import_component('comm', comm)
                    vardict.append(
                        (curr_prefix + '%s:' % comm.upper(), ''))
                    curr_prefix += prefix
                    avail = [cmm.is_installed(language=lang)
                             for lang in comm_lang_list]
                    vardict.append(
                        (curr_prefix + "Available for ",
                         sorted([comm_lang_list[i].upper()
                                 for i in range(len(avail))
                                 if avail[i]])))
                    vardict.append(
                        (curr_prefix + "Not Available for ",
                         sorted([comm_lang_list[i].upper()
                                 for i in range(len(avail))
                                 if not avail[i]])))
                    curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
            # Add verbose information
            if args.verbose:
                # Path variables
                path_vars = ['PATH', 'C_INCLUDE_PATH', 'INCLUDE',
                             'LIBRARY_PATH', 'LD_LIBRARY_PATH', 'LIB']
                vardict.append(('Environment paths:', ''))
                curr_prefix += prefix
                for k in path_vars:
                    if os.environ.get(k, ''):
                        vardict.append(
                            (curr_prefix + k, '\n%s%s'
                             % (curr_prefix + prefix,
                                ("\n" + curr_prefix + prefix).join(
                                    os.environ[k].split(os.pathsep)))))
                    else:
                        vardict.append((curr_prefix + k, ''))
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Environment variabless
                env_vars = ['CONDA_PREFIX', 'CONDA', 'SDKROOT', 'CC',
                            'CXX', 'FC', 'GFORTRAN', 'DISPLAY', 'CL',
                            '_CL_']
                if platform._is_win:  # pragma: windows
                    env_vars += ['VCPKG_ROOT']
                vardict.append(('Environment variables:', ''))
                curr_prefix += prefix
                for k in env_vars:
                    vardict.append(
                        (curr_prefix + k, os.environ.get(k, None)))
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Conda info
                if os.environ.get('CONDA_PREFIX', ''):
                    if platform._is_win:  # pragma: windows
                        out = tools.bytes2str(subprocess.check_output(
                            'conda info', shell=True)).strip()
                    else:
                        out = tools.bytes2str(subprocess.check_output(
                            ['conda', 'info'])).strip()
                    curr_prefix += prefix
                    vardict.append(
                        (curr_prefix + 'Conda Info:', "\n%s%s"
                         % (curr_prefix + prefix,
                            ("\n" + curr_prefix + prefix).join(
                                out.splitlines(False)))))
                    curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Configuration
                with open(config.usr_config_file, 'r') as fd:
                    contents = fd.read()
                vardict.append(
                    ('Configuration file:', '%s\n\t%s' % (
                        config.usr_config_file,
                        '\n\t'.join(contents.splitlines()))))
                # R and reticulate info
                Rdrv = import_component("model", "R")
                if Rdrv.is_installed():
                    env_reticulate = copy.deepcopy(os.environ)
                    env_reticulate['RETICULATE_PYTHON'] = sys.executable
                    # Stack size
                    out = Rdrv.run_executable(
                        ["-e", "Cstack_info()"]).strip()
                    vardict.append(
                        (curr_prefix + "R Cstack_info:", "\n%s%s"
                         % (curr_prefix + prefix,
                            ("\n" + curr_prefix + prefix).join(
                                out.splitlines(False)))))
                    # Compilation tools
                    interp = 'R'.join(
                        Rdrv.get_interpreter().rsplit('Rscript', 1))
                    vardict.append(
                        (curr_prefix + "R C Compiler:", ""))
                    curr_prefix += prefix
                    for x in ['CC', 'CFLAGS', 'CXX', 'CXXFLAGS']:
                        try:
                            out = tools.bytes2str(
                                subprocess.check_output(
                                    [interp, 'CMD', 'config', x],
                                    stderr=subprocess.STDOUT)).strip()
                        except subprocess.CalledProcessError:
                            out = 'ERROR (missing Rtools?)'
                        vardict.append(
                            (curr_prefix + x, "%s"
                             % ("\n" + curr_prefix + prefix).join(
                                 out.splitlines(False))))
                    curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                    # Session info
                    out = Rdrv.run_executable(
                        ["-e", "sessionInfo()"]).strip()
                    vardict.append(
                        (curr_prefix + "R sessionInfo:", "\n%s%s"
                         % (curr_prefix + prefix,
                            ("\n" + curr_prefix + prefix).join(
                                out.splitlines(False)))))
                    # Reticulate conda_list
                    if os.environ.get('CONDA_PREFIX', ''):
                        if platform._is_win:  # pragma: windows
                            out = tools.bytes2str(subprocess.check_output(
                                'conda info --json', shell=True)).strip()
                        else:
                            out = tools.bytes2str(subprocess.check_output(
                                ['conda', 'info', '--json'])).strip()
                        vardict.append(
                            (curr_prefix
                             + "conda info --json",
                             "\n%s%s" % (
                                 curr_prefix + prefix,
                                 ("\n" + curr_prefix + prefix).join(
                                     out.splitlines(False)))))
                        try:
                            out = Rdrv.run_executable(
                                ["-e", ("library(reticulate); "
                                        "reticulate::conda_list()")],
                                env=env_reticulate).strip()
                        except BaseException:  # pragma: debug
                            out = 'ERROR'
                        vardict.append(
                            (curr_prefix
                             + "R reticulate::conda_list():",
                             "\n%s%s" % (
                                 curr_prefix + prefix,
                                 ("\n" + curr_prefix + prefix).join(
                                     out.splitlines(False)))))
                    # Windows python versions
                    if platform._is_win:  # pragma: windows
                        out = Rdrv.run_executable(
                            ["-e",
                             ("library(reticulate); "
                              "reticulate::py_versions_windows()")],
                            env=env_reticulate).strip()
                        vardict.append(
                            (curr_prefix
                             + "R reticulate::py_versions_windows():",
                             "\n%s%s" % (
                                 curr_prefix + prefix,
                                 ("\n" + curr_prefix + prefix).join(
                                     out.splitlines(False)))))
                    # conda_binary
                    if platform._is_win and shutil.which('conda'):  # pragma: windows
                        out = Rdrv.run_executable(
                            ["-e",
                             ("library(reticulate); "
                              "conda <- reticulate:::conda_binary(\"auto\"); "
                              "system(paste(conda, \"info --json\"))")],
                            env=env_reticulate).strip()
                        vardict.append(
                            (curr_prefix
                             + "R reticulate::py_versions_windows():",
                             "\n%s%s" % (
                                 curr_prefix + prefix,
                                 ("\n" + curr_prefix + prefix).join(
                                     out.splitlines(False)))))
                    # Reticulate py_config
                    out = Rdrv.run_executable(
                        ["-e", ("library(reticulate); "
                                "reticulate::py_config()")],
                        env=env_reticulate).strip()
                    vardict.append(
                        (curr_prefix + "R reticulate::py_config():",
                         "\n%s%s" % (
                             curr_prefix + prefix,
                             ("\n" + curr_prefix + prefix).join(
                                 out.splitlines(False)))))
        finally:
            # Print things
            max_len = max(len(x[0]) for x in vardict)
            lines = []
            line_format = '%-' + str(max_len) + 's' + prefix + '%s'
            for k, v in vardict:
                lines.append(line_format % (k, v))
            msg = "yggdrasil info:\n%s" % '\n'.join(lines)
            if return_str:
                return msg
            logger.info(msg)

        
class validate_yaml(SubCommand):
    r"""Validate a set of or or more YAMLs defining an integration."""

    name = "validate"
    help = 'Validate a set of YAML specification files for an integration.'
    arguments = [
        (('yamlfile', ),
         {'nargs': '+',
          'help': 'One or more YAML specification files.'}),
        (('--model-only', ),
         {'action': 'store_true',
          'help': ('Validate a YAML containing an isolated model without '
                   'ensuring that it is part of a complete integration.')}),
        (('--model-submission', ),
         {'action': 'store_true',
          'help': ('Validate a YAML against the requirements for '
                   'submissions to the yggdrasil model repository.')})]

    @classmethod
    def func(cls, args):
        if args.model_submission:
            from yggdrasil.services import validate_model_submission
            validate_model_submission(args.yamlfile)
        else:
            from yggdrasil import yamlfile
            yamlfile.parse_yaml(args.yamlfile, model_only=args.model_only,
                                model_submission=args.model_submission)
        logger.info("Validation succesful.")


class yggcc(SubCommand):
    r"""Compile a program."""

    name = "compile"
    help = ("Compile a program from source files for use in an "
            "yggdrasil integration.")
    arguments = [
        (('source', ),
         {'nargs': '+',
          'help': ("One or more source files or the directory containing an "
                   "R package.")}),
        (('--language', ),
         {'default': None,
          'choices': ([None] + LANGUAGES_WITH_ALIASES.get('compiled', [])
                      + ['R', 'r']),
          'help': ("Language of the source code. If not provided, "
                   "the language will be determined from the "
                   "source extension.")}),
        (('--toolname', ),
         {'help': "Name of compilation tool that should be used"}),
        (('--flags', ),
         {'nargs': '*',
          'help': ("Additional flags that should be added to the compilation "
                   "command")}),
        (('--use-ccache', ),
         {'action': 'store_true', 'help': "Run compilation with ccache."}),
        (('--Rpkg-language', ),
         {'help': ("Language that R package is written in "
                   "(only used if the specified language is R).")})]

    @classmethod
    def func(cls, args):
        from yggdrasil.components import import_component
        from yggdrasil.constants import EXT2LANG
        if args.language is None:
            if (((len(args.source) == 1) and os.path.isdir(args.source[0])
                 and os.path.isdir(os.path.join(args.source[0], 'R')))):
                args.language = 'R'
            else:
                args.language = EXT2LANG[os.path.splitext(args.source[0])[-1]]
        drv = import_component('model', args.language)
        kws = {'toolname': args.toolname, 'flags': args.flags,
               'use_ccache': args.use_ccache}
        if (args.language in ['r', 'R']) and args.Rpkg_language:
            kws['language'] = args.Rpkg_language
        print("executable: %s" % drv.call_compiler(args.source, **kws))


class yggcompile(SubCommand):
    r"""Compile interface library/libraries."""

    name = "compile-deps"
    help = ("Compile yggdrasil dependency libraries. Existing "
            "libraries are first deleted.")
    arguments = [
        (('language', ),
         {'nargs': '*', 'default': ['all'],
          # 'choices': (['all'] + LANGUAGES_WITH_ALIASES.get('compiled', [])
          #             + LANGUAGES_WITH_ALIASES.get('compiled_dsl', [])),
          'help': ("One or more languages to compile dependencies "
                   "for.")}),
        (('--toolname', ),
         {'help': "Name of compilation tool that should be used"})]

    @classmethod
    def func(cls, args):
        from yggdrasil.components import import_component
        yggclean.func(args, verbose=False)
        error_on_missing = (not getattr(args, 'all_languages', False))
        missing = []
        for lang in args.language:
            import_component('model', lang).cleanup_dependencies()
            drv = import_component('model', lang)
            if ((hasattr(drv, 'compile_dependencies')
                 and (not getattr(drv, 'is_build_tool', False)))):
                if drv.is_installed():
                    drv.compile_dependencies(toolname=args.toolname)
                else:
                    missing.append(lang)
        if error_on_missing and missing:  # pragma: debug
            raise Exception(("One or more of the requested languages "
                             "are not fully installed for use with "
                             "yggdrasil: %s") % missing)


class yggclean(SubCommand):
    r"""Cleanup dependency files."""

    name = "clean"
    help = "Remove dependency libraries compiled by yggdrasil."
    arguments = [
        (('language', ),
         {'nargs': '*', 'default': ['all'],
          # 'choices': ['all'] + LANGUAGES_WITH_ALIASES.get('all', []),
          'help': ("One or more languages to clean up dependencies "
                   "for.")})]

    @classmethod
    def func(cls, args, verbose=True):
        from yggdrasil.components import import_component
        for lang in args.language:
            if lang in ['ipc', 'ipcs']:
                from yggdrasil.communication.IPCComm import ipcrm_queues
                ipcrm_queues()
            else:
                import_component('model', lang).cleanup_dependencies(
                    verbose=verbose)


class cc_toolname(SubCommand):
    r"""Output the name of the compiler used to compile C or C++ programs."""

    name = "compiler-tool"
    help = 'Get the compiler used for C/C++ programs.'
    arguments = [
        (('--cpp', ),
         {'action': 'store_true',
          'help': 'Get the compiler used for C++ programs.'}),
        (('--fullpath', ),
         {'action': 'store_true',
          'help': 'Get the full path to the tool exectuable.'})]

    @classmethod
    def parse_args(cls, *args, **kwargs):
        args = super(cc_toolname, cls).parse_args(*args, **kwargs)
        args.tool = 'compiler'
        args.toolname = None
        args.flags = False
        if args.cpp:
            args.language = 'cpp'
        else:
            args.language = 'c'
        return args

    @classmethod
    def func(cls, args, **kwargs):
        ygginfo.func(args, **kwargs)


class ld_toolname(cc_toolname):
    r"""Output the name of the linker used to compile C or C++ programs."""

    name = "linker-tool"
    help = 'Get the linker used for C/C++ programs.'
    arguments = [
        (('--cpp', ),
         {'action': 'store_true',
          'help': 'Get the linker used for C++ programs.'}),
        (('--fullpath', ),
         {'action': 'store_true',
          'help': 'Get the full path to the tool exectuable.'})]

    @classmethod
    def parse_args(cls, *args, **kwargs):
        args = super(ld_toolname, cls).parse_args(*args, **kwargs)
        args.tool = 'linker'
        return args


class cc_flags(cc_toolname):
    r"""Get the compiler flags necessary for including the interface
    library in a C or C++ program."""

    name = "compiler-flags"
    help = 'Get the compilation flags necessary for a C/C++ program.'
    arguments = [
        (('--cpp', ),
         {'action': 'store_true',
          'help': 'Get the compilation flags used for C++ programs.'}),
        (('--toolname', ),
         {'default': None,
          'help': 'Name of the tool that associated flags be returned for.'})]

    @classmethod
    def parse_args(cls, *args, **kwargs):
        args = super(cc_flags, cls).parse_args(*args, **kwargs)
        args.flags = True
        return args


class ld_flags(cc_toolname):
    r"""Get the linker flags necessary for including the interface
    library in a C or C++ program."""

    name = "linker-flags"
    help = 'Get the compilation flags necessary for a C/C++ program.'
    arguments = [
        (('--cpp', ),
         {'action': 'store_true',
          'help': 'Get the compilation flags used for C++ programs.'}),
        (('--toolname', ),
         {'default': None,
          'help': 'Name of the tool that associated flags be returned for.'})]

    @classmethod
    def parse_args(cls, *args, **kwargs):
        args = super(ld_flags, cls).parse_args(*args, **kwargs)
        args.tool = 'linker'
        args.flags = True
        return args


class ygginstall(SubCommand):
    r"""Call installation script."""

    name = "install"
    help = "Complete yggdrasil installation for one or more languages."

    @classmethod
    def get_parser(cls, **kwargs):
        from yggdrasil.languages import install_languages
        # TODO: Migrate
        return install_languages.update_argparser()

    @classmethod
    def func(cls, args):
        from yggdrasil.languages import install_languages
        languages = args.language
        if (((isinstance(languages, str) and (languages == 'all'))
             or (isinstance(languages, list) and ('all' in languages)))):
            languages = [x.lower() for x in
                         install_languages.get_language_directories()]
        for x in languages:
            install_languages.install_language(x, args=args)


class update_config(SubCommand):
    r"""Update the user config file for yggdrasil."""

    name = "config"
    help = 'Update the user config file.'
    arguments = (
        [(('languages', ),
          {'nargs': '*',
           # 'choices': ['all'] + LANGUAGES_WITH_ALIASES.get('all', []),
           'default': [],
           'help': 'One or more languages that should be configured.'}),
         (('--languages', ),
          {'nargs': '+', 'dest': 'languages_flag',
           # 'choices': ['all'] + LANGUAGES_WITH_ALIASES.get('all', []),
           'default': [],
           'help': 'One or more languages that should be configured.'}),
         (('--show-file', ),
          {'action': 'store_true',
           'help': 'Print the path to the config file without updating it.'}),
         (('--remove-file', ),
          {'action': 'store_true',
           'help': 'Remove the existing config file and return.'}),
         (('--overwrite', ),
          {'action': 'store_true',
           'help': 'Overwrite the existing file.'}),
         (('--disable-languages', ),
          {'nargs': '+', 'default': [],
           'choices': LANGUAGES_WITH_ALIASES.get('all', []),
           'help': 'One or more languages that should be disabled.'}),
         (('--enable-languages', ),
          {'nargs': '+', 'default': [],
           'choices': LANGUAGES_WITH_ALIASES.get('all', []),
           'help': 'One or more languages that should be enabled.'}),
         (('--quiet', '-q'),
          {'action': 'store_true',
           'help': 'Suppress output.'}),
         (('--allow-multiple-omp', ),
          {'action': 'store_true', 'default': None,
           'help': ('Have yggdrasil set the environment variable '
                    'KMP_DUPLICATE_LIB_OK to \'True\' during model runs '
                    'to disable errors resembling '
                    '"OMP: Error #15: Initializing libomp.dylib..." '
                    'that result from having multiple versions of OpenMP '
                    'loaded during runtime.')}),
         (('--dont-allow-multiple-omp', ),
          {'action': 'store_true', 'default': None,
           'help': ('Don\'t set the KMP_DUPLICATE_LIB_OK environment variable '
                    'when running models (see help for \'--allow-multiple-omp\' '
                    'for more information).')})]
        + [(('--%s-compiler' % k, ),
            {'help': ('Name or path to compiler that should be used to compile '
                      'models written in %s.' % k)})
           for k in LANGUAGES.get('compiled', [])]
        + [(('--%s-linker' % k, ),
            {'help': ('Name or path to linker that should be used to link '
                      'models written in %s.' % k)})
           for k in LANGUAGES.get('compiled', [])]
        + [(('--%s-archiver' % k, ),
            {'help': ('Name or path to archiver that should be used to create '
                      'static libraries for models written in %s.' % k)})
           for k in LANGUAGES.get('compiled', [])]
    )
    # TODO: Move these into the language directories?
    language_arguments = {
        'c': [
            ConditionalArgumentTuple(
                ('--vcpkg-dir', ),
                {'help': 'Directory containing the vcpkg installation.'},
                conditions={'os': ['Windows']}),
            ConditionalArgumentTuple(
                ('--macos-sdkroot', '--sdkroot'),
                {'help': ('The full path to the MacOS SDK that '
                          'should be used.')},
                conditions={'os': ['MacOS']})],
        'matlab': [
            (('--disable-matlab-engine-for-python', ),
             {'action': 'store_true', 'default': None,
              'dest': 'disable_engine',
              'help': 'Disable use of the Matlab engine for Python.'}),
            (('--enable-matlab-engine-for-python', ),
             {'action': 'store_true', 'default': None,
              'dest': 'enable_engine',
              'help': 'Enable use of the Matlab engine for Python.'}),
            (('--hide-matlab-libiomp', ),
             {'action': 'store_true', 'default': None,
              'help': ('Hide the version of libiomp installed by Matlab '
                       'by slightly changing the filename so that the '
                       'conda version of libomp is used instead. This '
                       'helps to solve the error "'
                       'OMP: Error #15: Initializing libomp.dylib..." '
                       'that can occur when using a conda environment. '
                       'The hidden file location will be set in the '
                       'configuration file and can be restored via the '
                       '\'--restore-matlab-libiomp\' option.')}),
            (('--restore-matlab-libiomp', ),
             {'action': 'store_true', 'default': None,
              'help': ('Restore the version of libiomp installed by Matlab. '
                       '(See help for \'--hide-matlab-libiomp\')')}),
        ],
        'osr': [
            (('--osr-repository-path', ),
             {'dest': 'repository',
              'help': 'The full path to the OpenSimRoot repository.'})]}
    opposite_arguments = [
        ('allow_multiple_omp', 'dont_allow_multiple_omp'),
        ('disable_engine', 'enable_engine'),
        ('hide_matlab_libiomp', 'restore_matlab_libiomp')]
        
    @classmethod
    def add_arguments(cls, parser, **kwargs):
        super(update_config, cls).add_arguments(parser, **kwargs)
        args = kwargs.get('args', None)
        if args is None:
            args = sys.argv[1:]
        if ('-h' in args) or ('--help' in args):
            args = [x for x in args if x not in ['-h', '--help']]
        preargs = parser.parse_known_args(args=args)[0]
        prelang = preargs.languages
        if preargs.languages_flag:
            prelang += preargs.languages_flag
        if (len(prelang) == 0) or ('all' in prelang):
            prelang = LANGUAGES.get('all', [])
        # TODO: The languages could be subparsers
        for k, v in cls.language_arguments.items():
            if k in prelang:
                cls.add_argument_to_parser(parser, v)

    @classmethod
    def func(cls, args):
        from yggdrasil import config
        if args.show_file:
            print('Config file located here: %s'
                  % config.usr_config_file)
        if args.remove_file and os.path.isfile(config.usr_config_file):
            os.remove(config.usr_config_file)
        if args.show_file or args.remove_file:
            return
        for x_true, x_false in cls.opposite_arguments:
            if getattr(args, x_false, None) is not None:
                assert(getattr(args, x_true, None) is None)
                setattr(args, x_true, not getattr(args, x_false))
            if hasattr(args, x_false):
                delattr(args, x_false)
        lang_kwargs = {}
        for k, v in cls.language_arguments.items():
            for v_args in v:
                name = v_args[1].get(
                    'dest',
                    v_args[0][0].lstrip('-').replace('-', '_'))
                if getattr(args, name, None) is not None:
                    lang_kwargs.setdefault(k, {})
                    lang_kwargs[k][name] = getattr(args, name)
        for x in ['compiler', 'linker', 'archiver']:
            for k in LANGUAGES.get('compiled', []):
                if getattr(args, '%s_%s' % (k, x), None):
                    lang_kwargs.setdefault(k, {})
                    lang_kwargs[k][x] = getattr(args, '%s_%s' % (k, x))
        config.update_language_config(
            args.languages, overwrite=args.overwrite,
            verbose=(not args.quiet),
            disable_languages=args.disable_languages,
            enable_languages=args.enable_languages,
            allow_multiple_omp=args.allow_multiple_omp,
            lang_kwargs=lang_kwargs)


class regen_metaschema(SubCommand):
    r"""Regenerate the yggdrasil metaschema."""

    name = "metaschema"
    help = "Regenerate the yggdrasil metaschema."

    @classmethod
    def func(cls, args):
        from yggdrasil import metaschema
        if os.path.isfile(metaschema._metaschema_fname):
            os.remove(metaschema._metaschema_fname)
        metaschema._metaschema = None
        metaschema._validator = None
        metaschema.get_metaschema()


class regen_schema(SubCommand):
    r"""Regenerate the yggdrasil schema."""

    name = "schema"
    help = "Regenerate the yggdrasil schema."
    arguments = [
        (('--only-constants', ),
         {'action': 'store_true',
          'help': ('Only update the constants.py file without updating '
                   'the schema.')})]

    @classmethod
    def func(cls, args):
        from yggdrasil import schema
        if not args.only_constants:
            if os.path.isfile(schema._schema_fname):
                os.remove(schema._schema_fname)
            schema.clear_schema()
            schema.init_schema()
        else:
            schema.update_constants()


class yggmodelform(SubCommand):
    r"""Save/print a JSON schema that can be used for generating a
    form for composing a model specification files."""

    name = "model-form-schema"
    help = ('Save/print the JSON schema for generating the model '
            'specification form.')
    arguments = [
        (('--file', ),
         {'help': 'Path to file where the schema should be saved.'})]

    @classmethod
    def func(cls, args):
        from yggdrasil.schema import get_model_form_schema
        out = get_model_form_schema(fname_dst=args.file, indent='    ')
        if not args.file:
            pprint.pprint(out)


class yggdevup(SubCommand):
    r"""Cleanup old libraries, re-install languages, and re-compile
    interface libraries following an update to the code (when doing
    development)."""

    name = "dev-update"
    help = ('Perform cleanup and reinitialization following an '
            'update to the code during development.')

    @classmethod
    def func(cls, args):
        yggclean(args=['all'])
        ygginstall(args=['all'])
        yggcompile(args=['all'])
        ygginfo(args=[])


class timing_plots(SubCommand):
    r"""Create performance plots using timing results."""

    name = "timing"
    help = "Create performance plots using timing results."
    arguments = [
        ArgumentSubparser(
            title='comparison', dest='comparison',
            description='Comparison plot that should be created.',
            parsers=[
                ArgumentParser(
                    name='commtype',
                    help='Compare different communication avenues.'),
                ArgumentParser(
                    name='language',
                    help='Compare different programming languages.'),
                ArgumentParser(
                    name='os',
                    help='Compare different operating systems.'),
                ArgumentParser(
                    name='python',
                    help='Compare different versions of Python.'),
                ArgumentParser(
                    name='lang2019',
                    help='Create plots from Lang (2019) paper.')])]

    @classmethod
    def func(cls, args):
        from yggdrasil import timing
        if args.comparison == 'lang2019':
            _lang_list = timing._lang_list
            _lang_list_nomatlab = copy.deepcopy(_lang_list)
            _lang_list_nomatlab.remove('matlab')
            timing.plot_scalings(compare='platform', python_ver='2.7')
            # All plots on Linux, no matlab
            timing.plot_scalings(compare='comm_type',
                                 platform='Linux',
                                 python_ver='2.7')
            timing.plot_scalings(compare='python_ver',
                                 platform='Linux')
            timing.plot_scalings(compare='language',
                                 platform='Linux',
                                 python_ver='2.7',
                                 compare_values=_lang_list_nomatlab)
            # Language comparision on MacOS, with matlab
            timing.plot_scalings(compare='language',
                                 platform='MacOS',
                                 python_ver='2.7',
                                 compare_values=_lang_list)
        else:
            if args.comparison == 'commtype':
                args.comparison = 'comm_type'
            elif args.comparison == 'os':
                args.comparison = 'platform'
            timing.plot_scalings(compare=args.comparison)


class generate_gha_workflow(SubCommand):
    r"""Re-generate the Github actions workflow yaml."""

    name = "gha"
    help = (
        "Generate a Github Actions (GHA) workflow yaml file from "
        "a version of the file that uses anchors (not supported by "
        "GHA as of 2021-01-14).")
    arguments = [
        (('--base', '--base-file'),
         {'help': (
             "Version of GHA workflow yaml that contains anchors.")}),
        (('--dest', ),
         {'help': (
             "Name of target GHA workflow yaml file.")}),
        (('--verbose', ),
         {'action': 'store_true', 'help': "Print yaml contents."})]

    @classmethod
    def func(cls, args, gitdir=None):
        import yaml
        from yggdrasil.metaschema.encoder import decode_yaml, encode_yaml
        from collections import OrderedDict

        class NoAliasDumper(yaml.SafeDumper):
            def ignore_aliases(self, data):
                return True
        if (args.base is None or args.dest is None) and (gitdir is None):
            try:
                gitdir = subprocess.check_output(
                    ["git", "rev-parse", "--git-dir"],
                    stderr=subprocess.PIPE).decode('utf-8').strip()
            except subprocess.CalledProcessError:
                return 1
        if args.base is None:
            args.base = os.path.join(gitdir, '..', 'utils',
                                     'test-install-base.yml')
        if args.dest is None:
            args.dest = os.path.join(gitdir, '..', '.github', 'workflows',
                                     'test-install.yml')
        with open(args.base, 'r') as fd:
            contents = decode_yaml(fd.read(), sorted_dict_type=OrderedDict)
            # contents = yaml.load(fd, Loader=yaml.SafeLoader)
        if args.verbose:
            pprint.pprint(contents)
        with open(args.dest, 'w') as fd:
            fd.write(('# DO NOT MODIFY THIS FILE, IT IS GENERATED.\n'
                      '# To make changes, modify \'%s\'\n'
                      '# and run \'ygggha\'\n')
                     % args.base)
            fd.write(encode_yaml(contents, sorted_dict_type=OrderedDict,
                                 Dumper=NoAliasDumper))
            # yaml.dump(contents, fd, Dumper=NoAliasDumper)


def rebuild_c_api():
    r"""Rebuild the C/C++ API."""
    ReplacementWarning('yggbuildapi_c', 'yggdrasil compile-deps c cpp')
    yggcompile(args=['c', 'cpp'] + sys.argv[1:])


def yggtime_comm():
    r"""Plot timing statistics comparing the different communication mechanisms."""
    ReplacementWarning('yggtime_comm', 'yggdrasil timing commtype')
    timing_plots(args=['commtype'] + sys.argv[1:])


def yggtime_lang():
    r"""Plot timing statistics comparing the different languages."""
    ReplacementWarning('yggtime_lang', 'yggdrasil timing language')
    timing_plots(args=['language'] + sys.argv[1:])


def yggtime_os():
    r"""Plot timing statistics comparing the different operating systems."""
    ReplacementWarning('yggtime_os', 'yggdrasil timing os')
    timing_plots(args=['os'] + sys.argv[1:])


def yggtime_py():
    r"""Plot timing statistics comparing the different versions of Python."""
    ReplacementWarning('yggtime_py', 'yggdrasil timing python')
    timing_plots(args=['python'] + sys.argv[1:])


def yggtime_paper():
    r"""Create plots for timing."""
    ReplacementWarning('yggtime_paper', 'yggdrasil timing lang2019')
    timing_plots(args=['lang2019'] + sys.argv[1:])


class main(SubCommand):
    r"""Runner for yggdrasil CLI."""

    name = "yggdrasil"
    help = (
        "Command line interface for the yggdrasil package.")
    arguments = []
    subcommands = [yggrun, ygginfo, validate_yaml,
                   yggcc, yggcompile, yggclean,
                   ygginstall, update_config,
                   regen_metaschema, regen_schema,
                   yggmodelform, yggdevup,
                   timing_plots, generate_gha_workflow,
                   integration_service_manager]

    @classmethod
    def get_parser(cls, **kwargs):
        from yggdrasil import __version__ as ver
        parser = super(main, cls).get_parser(**kwargs)
        parser.add_argument('--version', action='version',
                            version=('yggdrasil %s' % ver))
        subparsers = parser.add_subparsers(title='subcommands',
                                           dest='subcommand')
        parser._ygg_subparsers = {}
        for x in cls.subcommands:
            x.add_subparser(subparsers, args=kwargs.get('args', None))
            parser._ygg_subparsers[x.name] = x
        return parser

    @classmethod
    def parse_args(cls, parser, args=None, **kwargs):
        if args is None:
            args = sys.argv[1:]
        if isinstance(args, list) and ('test' in args):
            kwargs['allow_unknown'] = True
        args = super(main, cls).parse_args(parser, args=args, **kwargs)
        if args.subcommand:
            args = parser._ygg_subparsers[args.subcommand].parse_args(
                parser, args=args, **kwargs)
        return args

    @classmethod
    def func(cls, args):
        args.func(args)


if __name__ == '__main__':
    yggrun()
    sys.exit(0)



import pytest
from yggdrasil import platform
from tests.communication.test_CommBase import TestComm as base_class


class TestZMQComm(base_class):
    r"""Test for ZMQComm with non-default protocols and socket types."""

    test_send_recv_nolimit = None
    test_eof_no_close = None

    @pytest.fixture(scope="class", autouse=True)
    def commtype(self):
        r"""Communicator type being tested."""
        return "zmq"

    @pytest.fixture(scope="class", autouse=True)
    def use_async(self):
        r"""Whether communicator should be asynchronous or not."""
        return False

    @pytest.fixture(scope="class", autouse=True,
                    params=["inproc", "tcp", "ipc"])
    # Unsupported ['udp', 'pgm', 'epgm']
    def protocol(self, request):
        r"""Protocol that should be used."""
        return request.param

    @pytest.fixture(scope="class", autouse=True,
                    params=['PAIR', 'PUSH', 'PUB', 'ROUTER'])
    def socket_type(self, request, protocol):
        r"""Socket type that should be used."""
        if (((((protocol, request.param) == ('tcp', 'PAIR'))
              or ((protocol != 'tcp') and (request.param != 'PAIR')))
             and ((protocol, request.param) != ('inproc', 'PUSH')))):
            pytest.skip("Redundent combination.")
        return request.param

    @pytest.fixture(scope="class", autouse=True)
    def check_protocol(self, protocol):
        r"""Check that the protocol is installed."""
        if protocol == "ipc":
            from yggdrasil.communication import IPCComm
            if not IPCComm.IPCComm.is_installed(language='python'):
                pytest.skip("IPC not installed.")

    @pytest.fixture(scope="class")
    def sleep_after_connect(self, socket_type):
        r"""Indicates if sleep should occur after comm creation."""
        return (socket_type == 'ROUTER')

    @pytest.fixture(scope="class")
    def testing_options(self, python_class, options, protocol, socket_type):
        r"""Testing options."""
        out = python_class.get_testing_options(**options)
        out['kwargs'].update(protocol=protocol, socket_type=socket_type)
        return out

    def test_router_recv(self, socket_type, send_comm, recv_comm,
                         testing_options, do_send_recv):
        r"""Test router receipt of message from the dealer with an
        identity."""
        send_comm.protocol
        send_comm.host
        if socket_type != 'ROUTER':
            pytest.skip("Only valid for ROUTER socket_type")
        temp = send_comm
        send_comm = recv_comm
        recv_comm = temp
        send_comm.direction = 'send'
        recv_comm.direction = 'recv'
        do_send_recv(send_comm, recv_comm,
                     send_params={'kwargs': {
                         'identity': recv_comm.dealer_identity}})
        send_comm.direction = 'recv'
        recv_comm.direction = 'send'

    def test_get_socket_type_mate(self, run_once, python_module):
        r"""Test socket type matching."""
        for s, r in python_module._socket_type_pairs:
            assert(python_module.get_socket_type_mate(s) == r)
            assert(python_module.get_socket_type_mate(r) == s)
        with pytest.raises(ValueError):
            python_module.get_socket_type_mate('INVALID')

    def test_format_address(self, run_once, python_module):
        r"""Test format/parse of address."""
        protocol = 'tcp'
        host = '127.0.0.1'
        port = 5555
        address = python_module.format_address(protocol, host, port)
        result = python_module.parse_address(address)
        assert(result['protocol'] == protocol)
        assert(result['host'] == host)
        assert(result['port'] == port)
        with pytest.raises(ValueError):
            python_module.parse_address('INVALID')
        with pytest.raises(ValueError):
            python_module.parse_address('INVALID://')

    def test_invalid_protocol(self, run_once, commtype):
        r"""Test raise of an error in the event of an invalid protocol."""
        from yggdrasil.communication import new_comm
        with pytest.raises(ValueError):
            new_comm('test_invalid_protocol', commtype=commtype,
                     protocol='invalid')

    @pytest.mark.skipif(platform._is_mac, reason="Testing on MacOS")
    @pytest.mark.skipif(platform._is_win, reason="Testing on Windows")
    def test_error_on_send_open_twice(self, run_once, python_module,
                                      python_class, close_comm):
        r"""Test creation of the same send socket twice for an error."""
        from yggdrasil.communication import new_comm
        import zmq
        for s, r in python_module._socket_type_pairs:
            # Send comm
            name1 = 'test_%s' % s
            comm1 = new_comm(name1 + '_1', commtype='zmq', socket_type=s,
                             dont_open=True, socket_action='bind')
            with pytest.raises(zmq.ZMQError):
                python_class(name1 + '_2', socket_type=s,
                             address=comm1.opp_address, socket_action='bind')
            close_comm(comm1)


class TestZMQCommClient(TestZMQComm):
    r"""Test for ZMQComm communication class for client/server."""

    test_drain_messages = None

    @pytest.fixture(scope="class", autouse=True, params=['tcp', 'ipc'])
    def protocol(self, request):
        r"""Protocol that should be used."""
        return request.param

    @pytest.fixture(scope="class", autouse=True, params=['PAIR'])
    def socket_type(self, request, protocol):
        r"""Socket type that should be used."""
        return request.param

    @pytest.fixture(scope="class")
    def testing_options(self, python_class, options, protocol, socket_type):
        r"""Testing options."""
        out = python_class.get_testing_options(**options)
        out['kwargs'].update(protocol=protocol, socket_type=socket_type,
                             is_client=True)
        return out


class TestZMQCommREQ(TestZMQComm):
    r"""Test for ZMQComm communication class with REP/REQ socket."""

    @pytest.fixture(scope="class", autouse=True, params=['tcp'])
    def protocol(self, request):
        r"""Protocol that should be used."""
        return request.param

    @pytest.fixture(scope="class", autouse=True, params=['REQ'])
    def socket_type(self, request, protocol):
        r"""Socket type that should be used."""
        return request.param

    test_send_recv_condition = None

    def test_send_recv_filter_eof(self, filtered_comms, do_send_recv,
                                  send_comm, recv_comm, polling_interval):
        r"""Test send/recv of EOF with filter."""
        with pytest.raises(RuntimeError):
            do_send_recv(send_comm, recv_comm,
                         send_params={'method': 'send_eof'},
                         recv_params={'skip_wait': True,
                                      'timeout': polling_interval,
                                      'flag': False})

    def test_send_recv_filter_pass(self, filtered_comms, do_send_recv,
                                   send_comm, recv_comm, msg_filter_pass):
        r"""Test send/recv with filter that passes both messages."""
        with pytest.raises(RuntimeError):
            do_send_recv(send_comm, recv_comm, msg_filter_pass)
        
    def test_send_recv_filter_send_filter(self, filtered_comms, do_send_recv,
                                          send_comm, recv_comm,
                                          msg_filter_send, polling_interval):
        r"""Test send/recv with filter that blocks send."""
        with pytest.raises(RuntimeError):
            do_send_recv(send_comm, recv_comm, msg_filter_send,
                         recv_params={'message': recv_comm.empty_obj_recv,
                                      'timeout': polling_interval,
                                      'skip_wait': True})
        
    def test_send_recv_filter_recv_filter(self, filtered_comms, do_send_recv,
                                          send_comm, recv_comm,
                                          msg_filter_recv, polling_interval):
        r"""Test send/recv with filter that blocks recv."""
        with pytest.raises(RuntimeError):
            do_send_recv(send_comm, recv_comm, msg_filter_recv,
                         recv_params={'message': recv_comm.empty_obj_recv,
                                      'timeout': 10 * polling_interval,
                                      'skip_wait': True})



import pytest
from tests.metaschema.datatypes.test_JSONObjectMetaschemaType import (
    TestJSONObjectMetaschemaType as base_class)
import copy
# from yggdrasil.metaschema.datatypes import compare_schema


class TestSchemaMetaschemaType(base_class):
    r"""Test class for SchemaMetaschemaType class."""

    _mod = 'yggdrasil.metaschema.datatypes.SchemaMetaschemaType'
    _cls = 'SchemaMetaschemaType'

    @pytest.fixture(scope="class")
    def value(self, fulldef):
        r"""dict: Test value."""
        return dict(copy.deepcopy(fulldef), type='object')
    
    @pytest.fixture(scope="class")
    def fulldef(self):
        r"""dict: Full type definitions."""
        return {'type': 'schema'}

    @pytest.fixture(scope="class")
    def typedef_base(self):
        r"""dict: Base type definition."""
        return {'type': 'schema'}
    
    @pytest.fixture(scope="class")
    def valid_encoded(self, fulldef):
        r"""list: Encoded objects that are valid under this type."""
        return [fulldef]
    
    @pytest.fixture(scope="class")
    def valid_decoded(self, value):
        r"""list: Objects that are valid under this type."""
        return [value]
    
    @pytest.fixture(scope="class")
    def invalid_encoded(self):
        r"""list: Encoded objects that are invalid under this type."""
        return [{}]

    @pytest.fixture(scope="class")
    def invalid_decoded(self):
        r"""list: Objects that are invalid under this type."""
        return [{}]
        
    @pytest.fixture(scope="class")
    def compatible_objects(self, value):
        r"""list: Objects that are compatible with this type."""
        return [(value, value, None)]

    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return [('float', {'type': 'float',
                           'precision': 64}),
                ({'units': 'g'}, {'units': 'g',
                                  'type': 'scalar',
                                  'subtype': 'float',
                                  'precision': 64}),
                ({'title': 'a'}, {'title': 'a'}),
                ({'title': 'a', 'units': 'g'},
                 {'title': 'a', 'units': 'g',
                  'type': 'scalar', 'subtype': 'float',
                  'precision': 64}),
                ({}, {})]

    # @classmethod
    # def assert_result_equal(cls, x, y):
    #     r"""Assert that serialized/deserialized objects equal."""
    #     compare_schema(x, y)



"""Testing things."""
import pytest
import os
import sys
import uuid
import importlib
import copy
import types
import functools
import subprocess
from yggdrasil import platform


# Test data
data_dir = os.path.join(os.path.dirname(__file__), 'data')
data_list = [
    ('txt', 'ascii_file.txt'),
    ('table', 'ascii_table.txt')]
data = {k: os.path.join(data_dir, v) for k, v in data_list}


def get_timeout_args(args0=None):
    r"""Determine which arguments should be passed to the timeout subprocess.

    Args:
        args0 (list, optional): Initial set of arguments. Defaults to sys.argv[1:].

    Returns:
        dict: Argument information.

    """
    if args0 is None:
        args0 = sys.argv[1:]
    args_preserve_path = ['-c']
    args_remove = ['--clear-cache']
    args_remove_value = []
    args_add = ['--cov-append']
    if int(pytest.__version__.split('.')[0]) >= 6:
        args_add.append('--import-mode=importlib')
    args_remove_value_match = tuple([k + '=' for k in args_remove_value])
    out = {'args': [], 'rootdir': None}
    args = out['args']
    i = 0
    for i in range(1, len(args0)):
        v = args0[i]
        if v in args_remove:  # pragma: testing
            pass
        elif v in args_remove_value:  # pragma: testing
            i += 1
        elif v.startswith(args_remove_value_match):  # pragma: testing
            pass
        elif v in args_preserve_path:
            args.append(v)
            i += 1
            args.append(args0[i])
        elif os.path.isfile(v) or os.path.isdir(v):
            pass
        else:
            if v.startswith('--rootdir='):
                out['rootdir'] = v.split('=')[-1]
            elif v == '--rootdir':  # pragma: testing
                out['rootdir'] = args0[i + 1]
                i += 1
            args.append(v)
        i += 1
    for v in args_add:
        if '=' in v:
            if not any(vv.startswith(v.split('=')[0]) for vv in args):
                args.append(v)
        elif v not in args:  # pragma: testing
            args.append(v)
    return out


def timeout_decorator(*args, allow_arguments=False, **kwargs):
    r"""Patch for pytest timeout on windows to allow pytest to cache.

    Args:
        *args: Arguments are passed on to the pytest.mark.timeout decorator.
        **kwargs: Keyword arguments are passed on to the pytest.mark.timeout
            decorator.
        allow_arguments (bool, optional): If True, decoration with the timeout
            decorator will allow arguments to be passed to the test function.
            Defaults to False.

    Source:
        https://stackoverflow.com/questions/21827874/timeout-a-function-windows/
            48980413#48980413


    """
    env_flag = 'YGG_IN_TIMEOUT_PROCESS'
    if platform._is_win:
        kwargs['method'] = 'thread'
    method = kwargs.get('method', 'signal')
    pytest_deco = pytest.mark.timeout(*args, **kwargs)
    if (((method == 'thread') and (not os.environ.get(env_flag, False))
         and (not allow_arguments))):
        arginfo = get_timeout_args()
        testargs = arginfo['args']
        rootdir = arginfo['rootdir']
        
        def deco(func):
            if getattr(func, '_timeout_wrapped', False):
                return func
            if isinstance(func, type):
                for k in dir(func):
                    v = getattr(func, k)
                    if k.startswith('test_') and isinstance(v,
                                                            types.MethodType):
                        setattr(func, k, deco(v))  # pragma: testing
                func._timeout_wrapped = True
                return func

            @functools.wraps(func)
            def wrapper(*args, **kwargs):  # pragma: testing
                testname = os.environ['PYTEST_CURRENT_TEST'].split()[0]
                if rootdir:
                    testname = os.path.join(rootdir, testname)
                max_args = testname.count('::') - 1
                if (len(args) > max_args) or kwargs:  # pragma: debug
                    raise Exception("Arguments not compatible with forked "
                                    "timeout, add 'allow_arguments=True' to "
                                    "the decorator")
                env = copy.deepcopy(os.environ)
                env[env_flag] = '1'
                out = None
                try:
                    out = subprocess.run(['pytest'] + testargs + [testname],
                                         env=env, check=True,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
                except subprocess.CalledProcessError as e:  # pragma: debug
                    out = e
                print(out.stdout.decode('utf-8'))
                print(out.stderr.decode('utf-8'), file=sys.stderr)
                if b'========= 1 skipped in' in out.stdout:
                    pytest.skip('')
                if out.returncode != 0:  # pragma: debug
                    raise RuntimeError("Error in test subprocess. "
                                       "See above output.")
            wrapper._timeout_wrapped = True
            return wrapper
        return deco
    else:
        return pytest_deco


@pytest.mark.usefixtures("verify_count_threads",
                         "verify_count_comms",
                         "verify_count_fds")
class TestBase:
    r"""Base class for tests."""
    
    @pytest.fixture(scope="class", autouse=True)
    def reset_test(self):
        self.__class__._first_test = True

    @pytest.fixture(autouse=True)
    def first_test(self):
        yield self.__class__._first_test
        self.__class__._first_test = False

    @pytest.fixture(scope="class")
    def global_uuid(self):
        r"""Unique ID for test."""
        return str(uuid.uuid4()).split('-')[0]

    @pytest.fixture
    def uuid(self):
        r"""Unique ID for test."""
        return str(uuid.uuid4()).split('-')[0]


class TestClassBase(TestBase):
    r"""Base class for testing classes."""

    @pytest.fixture(scope="class", autouse=True)
    def reset_test(self, module_name, class_name):
        self.__class__._first_test = True

    @pytest.fixture(scope="class")
    def module_name(self):
        r"""Name of the module containing the class being tested."""
        return self._mod
    
    @pytest.fixture(scope="class")
    def class_name(self):
        r"""Name of class that will be tested."""
        return self._cls

    @pytest.fixture(scope="class", autouse=True)
    def python_module(self, module_name):
        r"""Python module that is being tested."""
        return importlib.import_module(module_name)

    @pytest.fixture(scope="class", autouse=True)
    def python_class(self, python_module, class_name):
        r"""Python class that is being tested."""
        return getattr(python_module, class_name)

    @pytest.fixture(scope="class")
    def python_class_installed(self):
        r"""bool: True if the python class is installed."""
        return True

    @pytest.fixture(scope="class")
    def is_installed(self, class_name, python_class_installed):
        r"""Skip unless the component is installed."""
        if not python_class_installed:
            pytest.skip(f"{class_name} not installed")
    
    @pytest.fixture(scope="class")
    def is_not_installed(self, class_name, python_class_installed):
        r"""Skip unless the language is NOT installed."""
        if python_class_installed:
            pytest.skip(f"{class_name} installed")

    @pytest.fixture
    def instance_args(self):
        r"""Arguments for a new instance of the tested class."""
        return tuple([])

    @pytest.fixture
    def instance_kwargs(self):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict()

    @pytest.fixture
    def instance(self, python_class, instance_args, instance_kwargs,
                 verify_count_threads, verify_count_comms, verify_count_fds,
                 is_installed):
        r"""New instance of the python class for testing."""
        out = python_class(*instance_args, **instance_kwargs)
        yield out
        del out


class TestComponentBase(TestClassBase):

    _component_type = None

    @pytest.fixture(scope="class", autouse=True)
    def reset_test(self, component_subtype):
        self.__class__._first_test = True

    @pytest.fixture(scope="class", autouse=True)
    def component_type(self):
        r"""Type of component being tested."""
        return self._component_type

    @pytest.fixture(scope="class", autouse=True, params=[])
    def component_subtype(self, request):
        r"""Subtype of component being tested."""
        return request.param

    @pytest.fixture(scope="class")
    def module_name(self, python_class):
        r"""Name of the module containing the class being tested."""
        return importlib.import_module(python_class.__module__)
    
    @pytest.fixture(scope="class")
    def class_name(self, python_class):
        r"""Name of class that will be tested."""
        return python_class.__name__
    
    @pytest.fixture(scope="class", autouse=True)
    def python_module(self, python_class):
        r"""Python module that is being tested."""
        return importlib.import_module(python_class.__module__)

    @pytest.fixture(scope="class", autouse=True)
    def python_class(self, component_type, component_subtype):
        r"""Python class that is being tested."""
        from yggdrasil.components import import_component
        cls = import_component(component_type, component_subtype)
        return cls

    @pytest.fixture(scope="class", autouse=True)
    def options(self):
        r"""Arguments that should be provided when getting testing options."""
        return {}

    @pytest.fixture(scope="class")
    def testing_options(self, python_class, options):
        r"""Testing options."""
        if 'explicit_testing_options' in options:
            return copy.deepcopy(options['explicit_testing_options'])
        return python_class.get_testing_options(**options)

    @pytest.fixture
    def instance_kwargs(self, testing_options):
        r"""Keyword arguments for a new instance of the tested class."""
        return copy.deepcopy(testing_options.get('kwargs', {}))



import pytest
try:
    import trimesh
except ImportError:  # pragma: debug
    trimesh = None
from tests.demos import DemoTstBase as base_class


@pytest.mark.skipif(trimesh is None, reason="Trimesh is not installed")
class TestCiS2021HackathonDemo(base_class):

    @pytest.fixture(scope="class")
    def demo_name(self):
        r"""str: Name of demo being tested."""
        return 'CiS2021-hackathon'

    runs = {'light_v0': ('yamls/light_v0_python.yml', 'yamls/connections_v0.yml'),
            'shoot_v0': ('yamls/shoot_v0.yml', ),
            'shoot_v1': ('yamls/shoot_v1.yml', 'yamls/light_v0_python.yml',
                         'yamls/connections_v1.yml'),
            'shoot_v2': ('yamls/light_v1_python.yml', 'yamls/shoot_v2.yml'),
            'shoot_v2_split': ('yamls/light_v1_python.yml',
                               'yamls/shoot_v2_split.yml'),
            'shoot_v2_copies': ('yamls/light_v2_python.yml',
                                'yamls/shoot_v2_copies.yml'),
            'roots_v0': ('yamls/roots_v0.yml', ),
            'timesync': ('yamls/roots_v1.yml', 'yamls/shoot_v3.yml',
                         'yamls/timesync.yml', 'yamls/light_v1_python.yml')}



import pytest
from tests.metaschema.datatypes.test_MetaschemaType import (
    TestMetaschemaType as base_class)
from yggdrasil.metaschema.datatypes.ClassMetaschemaType import ExampleClass
import os
import tempfile


class TestClassMetaschemaType(base_class):
    r"""Test class for ClassMetaschemaType class."""

    _mod = 'yggdrasil.metaschema.datatypes.ClassMetaschemaType'
    _cls = 'ClassMetaschemaType'

    @pytest.fixture(scope="class")
    def value(self):
        r"""type: Test class."""
        return ExampleClass
    
    @pytest.fixture(scope="class")
    def valid_encoded(self, python_class, typedef_base):
        r"""list: Encoded objects that are valid under this type."""
        return [dict(typedef_base,
                     type=python_class.name)]
    
    @pytest.fixture(scope="class")
    def valid_decoded(self, value):
        r"""list: Objects that are valid under this type."""
        return [value, object]
    
    @pytest.fixture(scope="class")
    def invalid_decoded(self):
        r"""list: Objects that are invalid under this type."""
        return ['hello']

    @pytest.fixture(scope="class")
    def compatible_objects(self, value):
        r"""list: Objects that are compatible with this type."""
        return [(value, value, None)]

    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return [(None, None),
                ('yggdrasil.metaschema.datatypes.ClassMetaschemaType:'
                 'ExampleClass', ExampleClass)]

    def test_module_file(self, python_class):
        r"""Test decoding data that includes the full path to the module file."""
        _tmpclassfile = os.path.join(tempfile.gettempdir(), 'test_class.py')
        with open(_tmpclassfile, 'w') as fd:
            fd.write('class ValidClass2(object):\n    pass\n')
        out = python_class.decode_data('%s:ValidClass2' % _tmpclassfile, None)
        assert(out.__name__ == 'ValidClass2')

    def test_decode_data_errors(self, python_class):
        r"""Test errors in decode_data."""
        with pytest.raises(ValueError):
            python_class.decode_data('hello', None)
        with pytest.raises(ImportError):
            python_class.decode_data('invalid:invalid', None)
        with pytest.raises(AttributeError):
            python_class.decode_data('yggdrasil:invalid', None)



import os
from yggdrasil import tools
from yggdrasil.runner import run
tyk_dir = 'tyk3'

try:
    # Change to the solution directory
    old_dir = os.getcwd()
    if not old_dir.endswith(tyk_dir):
        os.chdir(os.path.join('solutions', tyk_dir))
    if not os.path.isdir('output'):
        os.mkdir('output')

    # Part 1: light-to-weather
    tools.display_source('yamls/light_to_weather.yml', number_lines=True)
    run(['yamls/light_to_weather.yml', 'yamls/light_v0_python.yml'],
        production_run=True)

    # Part 2: shoot-to-co2
    tools.display_source('yamls/shoot_to_co2.yml', number_lines=True)
    run(['yamls/shoot_to_co2.yml', 'yamls/shoot_v1.yml', 'yamls/co2.yml'],
        production_run=True)

    # Part 3: shoot-to-light&co2
    tools.display_source('yamls/shoot_to_light_and_co2.yml', number_lines=True)
    run(['yamls/shoot_to_light_and_co2.yml', 'yamls/shoot_v1.yml',
         'yamls/co2.yml', 'yamls/light_v0_python.yml'], production_run=True)

finally:
    os.chdir(old_dir)



import pytest
from tests.metaschema.properties.test_MetaschemaProperty import (
    TestMetaschemaProperty as base_class)


class TestItemsMetaschemaProperty(base_class):
    r"""Test class for ItemsMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.JSONArrayMetaschemaProperties'
    _cls = 'ItemsMetaschemaProperty'
    
    @pytest.fixture(scope="class")
    def nele(self):
        r"""Number of elements in test arrays."""
        return 3

    @pytest.fixture(scope="class")
    def valid_value(self, nele):
        r"""Base class for valid arrays."""
        return [int(i) for i in range(nele)]

    @pytest.fixture(scope="class")
    def valid_sing(self):
        r"""Valid type for single element."""
        return {'type': 'int'}

    @pytest.fixture(scope="class")
    def valid_mult(self, nele, valid_sing):
        r"""Valid type for multiple elements."""
        return [valid_sing for _ in range(nele)]

    @pytest.fixture(scope="class")
    def invalid_sing(self):
        r"""Invalid type for single element."""
        return {'type': 'float'}

    @pytest.fixture(scope="class")
    def invalid_mult(self, nele, invalid_sing):
        r"""Inalid type for multiple elements."""
        return [invalid_sing for _ in range(nele)]

    @pytest.fixture(scope="class")
    def valid(self, nele, valid_value, valid_sing, valid_mult):
        r"""Objects that are valid."""
        return [(valid_value, valid_sing),
                (valid_value, valid_mult),
                ([int(i) for i in range(nele - 1)], valid_sing)]

    @pytest.fixture(scope="class")
    def invalid(self, nele, valid_sing, valid_mult):
        r"""Objects that are invalid."""
        return [([float(i) for i in range(nele)], valid_sing),
                ([float(i) for i in range(nele)], valid_mult)]

    @pytest.fixture(scope="class")
    def valid_compare(self, nele, valid_sing, valid_mult):
        r"""Objects that successfully compare."""
        return [(valid_sing, valid_sing),
                (valid_sing, valid_mult),
                (valid_mult, valid_sing),
                (valid_mult, valid_mult)]

    @pytest.fixture(scope="class")
    def invalid_compare(self, valid_sing, invalid_sing,
                        valid_mult, invalid_mult):
        r"""Objects that do not successfully compare."""
        return [(valid_sing, invalid_sing),
                (valid_sing, invalid_mult),
                (valid_mult, invalid_sing),
                (valid_mult, invalid_mult),
                (1, 1),
                (valid_mult, valid_mult[:-1])]
